import re
import string


def normalize_span(s):
    if not s:
        return ""

    # Lower text and remove punctuation, articles and extra whitespace.
    def remove_articles(text):
        regex = re.compile(r"\b(a|an|the)\b", re.UNICODE)
        return re.sub(regex, " ", text)

    def white_space_fix(text):
        return " ".join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return "".join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(s))))



from collections import namedtuple
from typing import List

import tensorflow as tf
from tokenizers import BertWordPieceTokenizer
from transformers_keras.common.abc_dataset import AbstractDataset

UnsupervisedSimCSEExample = namedtuple(
    "UnsupervisedSimCSEExample",
    [
        "tokens",
        "input_ids",
        "segment_ids",
        "attention_mask",
    ],
)
SupervisedSimCSEExample = namedtuple(
    "SupervisedSimCSEExample",
    [
        "tokens",
        "input_ids",
        "segment_ids",
        "attention_mask",
        "pos_tokens",
        "pos_input_ids",
        "pos_segment_ids",
        "pos_attention_mask",
    ],
)

HardNegativeSimCSEExample = namedtuple(
    "HardNegativeSimCSEExample",
    [
        "tokens",
        "input_ids",
        "segment_ids",
        "attention_mask",
        "pos_tokens",
        "pos_input_ids",
        "pos_segment_ids",
        "pos_attention_mask",
        "neg_tokens",
        "neg_input_ids",
        "neg_segment_ids",
        "neg_attention_mask",
    ],
)


class UnsupervisedSimCSEDataset(AbstractDataset):
    """Dataset builder for unsupervised SimCSE"""

    @classmethod
    def _filter(cls, dataset, max_sequence_length=512, **kwargs):
        dataset = dataset.filter(lambda a, b, c: tf.size(a) <= max_sequence_length)
        return dataset

    @classmethod
    def _bucket_padding(
        cls,
        dataset,
        batch_size=32,
        pad_id=0,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        drop_remainder=False,
        **kwargs
    ):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = cls._bucketing(
            dataset,
            element_length_func=lambda a, b, c: tf.size(a),
            padded_shapes=([None, ], [None, ], [None, ]),
            padding_values=(pad_id, pad_id, pad_id),
            batch_size=batch_size,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs,
        )
        # fmt: on
        return dataset

    @classmethod
    def _batch_padding(cls, dataset, batch_size=32, pad_id=0, drop_remainder=False, **kwargs):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([None, ], [None, ], [None, ]),
            padding_values=(pad_id, pad_id, pad_id),
            drop_remainder=drop_remainder,
        )
        # fmt: on
        return dataset

    @classmethod
    def _fixed_padding(cls, dataset, batch_size=32, pad_id=0, max_sequence_length=512, drop_remainder=False, **kwargs):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        maxlen = tf.constant(max_sequence_length, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([None, ], [None, ], [None, ]),
            padding_values=(pad_id, pad_id, pad_id),
            drop_remainder=drop_remainder,
        )
        # fmt: on
        return dataset

    @classmethod
    def _zip_dataset(cls, examples: List[UnsupervisedSimCSEExample], **kwargs):
        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset([e.input_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.segment_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.attention_mask for e in examples], dtype=tf.int32),
            )
        )
        return dataset

    @classmethod
    def _to_dict(cls, dataset, **kwargs):
        dataset = dataset.map(
            lambda x, y, z: ({"input_ids": x, "segment_ids": y, "attention_mask": z}, None),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _parse_instances_to_examples(
        cls, instances, tokenizer: BertWordPieceTokenizer = None, vocab_file=None, **kwargs
    ):
        assert tokenizer or vocab_file, "`tokenizer` or `vocab_file` must be provided."
        if tokenizer is None:
            tokenizer = BertWordPieceTokenizer.from_file(
                vocab_file,
                lowercase=kwargs.get("do_lower_case", True),
            )
        examples = []
        for instance in instances:
            sequence = instance[kwargs.get("sequence_key", "sequence")]
            seq_encoding = tokenizer.encode(sequence)
            example = UnsupervisedSimCSEExample(
                tokens=seq_encoding.tokens,
                input_ids=seq_encoding.ids,
                segment_ids=seq_encoding.type_ids,
                attention_mask=seq_encoding.attention_mask,
            )
            examples.append(example)
        return examples

    @classmethod
    def _parse_tfrecord(cls, dataset, **kwargs):
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
            ),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _example_to_tfrecord(cls, example, **kwargs):
        feature = {
            "input_ids": cls._int64_feature([int(x) for x in example.input_ids]),
            "segment_ids": cls._int64_feature([int(x) for x in example.segment_ids]),
            "attention_mask": cls._int64_feature([int(x) for x in example.attention_mask]),
        }
        return tf.train.Example(features=tf.train.Features(feature=feature))


class SupervisedSimCSEDataset(AbstractDataset):
    """Dataset builder for supervised SimCSE"""

    @classmethod
    def _filter(cls, dataset, max_sequence_length=512, **kwargs):
        dataset = dataset.filter(
            lambda a, b, c, e, f, g: tf.logical_and(
                tf.size(a) <= max_sequence_length, tf.size(e) <= max_sequence_length
            )
        )
        return dataset

    @classmethod
    def _bucket_padding(
        cls,
        dataset,
        batch_size=32,
        pad_id=0,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        drop_remainder=False,
        **kwargs
    ):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = cls._bucketing(
            dataset,
            element_length_func=lambda a, b, c, e, f, g: tf.maximum(tf.size(a), tf.size(e)),
            padded_shapes=([None,], [None,], [None,], [None,], [None,], [None,]),
            padding_values=(pad_id, pad_id, pad_id, pad_id, pad_id, pad_id),
            batch_size=batch_size,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs,
        )
        # fmt: on
        return dataset

    @classmethod
    def _batch_padding(cls, dataset, batch_size=32, pad_id=0, drop_remainder=False, **kwargs):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([None,], [None,], [None,], [None,], [None,], [None,]),
            padding_values=(pad_id, pad_id, pad_id, pad_id, pad_id, pad_id),
            drop_remainder=drop_remainder,
        )
        # fmt: on
        return dataset

    @classmethod
    def _fixed_padding(cls, dataset, batch_size=32, pad_id=0, max_sequence_length=512, drop_remainder=False, **kwargs):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        maxlen = tf.constant(max_sequence_length, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([maxlen,], [maxlen,], [maxlen,], [maxlen,], [maxlen,], [maxlen,]),
            padding_values=(pad_id, pad_id, pad_id, pad_id, pad_id, pad_id),
            drop_remainder=drop_remainder,
        )
        # fmt: on
        return dataset

    @classmethod
    def _zip_dataset(cls, examples: List[SupervisedSimCSEExample], **kwargs):
        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset([e.input_ids for e in examples]),
                _to_dataset([e.segment_ids for e in examples]),
                _to_dataset([e.attention_mask for e in examples]),
                _to_dataset([e.pos_input_ids for e in examples]),
                _to_dataset([e.pos_segment_ids for e in examples]),
                _to_dataset([e.pos_attention_mask for e in examples]),
            )
        )
        return dataset

    @classmethod
    def _to_dict(cls, dataset, **kwargs):
        dataset = dataset.map(
            lambda a, b, c, e, f, g: (
                {
                    "input_ids": a,
                    "segment_ids": b,
                    "attention_mask": c,
                    "pos_input_ids": e,
                    "pos_segment_ids": f,
                    "pos_attention_mask": g,
                },
                None,
            ),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _parse_instances_to_examples(
        cls, instances, tokenizer: BertWordPieceTokenizer = None, vocab_file=None, **kwargs
    ):
        assert tokenizer or vocab_file, "`tokenizer` or `vocab_file` must be provided."
        if tokenizer is None:
            tokenizer = BertWordPieceTokenizer.from_file(
                vocab_file,
                lowercase=kwargs.get("do_lower_case", True),
            )
        examples = []
        for instance in instances:
            sequence = instance[kwargs.get("sequence_key", "sequence")]
            seq_encoding = tokenizer.encode(sequence)
            pos_sequence = instance[kwargs.get("pos_sequence_key", "pos_sequence")]
            pos_encoding = tokenizer.encode(pos_sequence)
            example = SupervisedSimCSEExample(
                tokens=seq_encoding.tokens,
                input_ids=seq_encoding.ids,
                segment_ids=seq_encoding.type_ids,
                attention_mask=seq_encoding.attention_mask,
                pos_tokens=pos_encoding.tokens,
                pos_input_ids=pos_encoding.ids,
                pos_segment_ids=pos_encoding.type_ids,
                pos_attention_mask=pos_encoding.attention_mask,
            )
            examples.append(example)
        return examples

    @classmethod
    def _parse_tfrecord(cls, dataset, **kwargs):
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "pos_input_ids": tf.io.VarLenFeature(tf.int64),
            "pos_segment_ids": tf.io.VarLenFeature(tf.int64),
            "pos_attention_mask": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["pos_input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["pos_segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["pos_attention_mask"]), tf.int32),
            ),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _example_to_tfrecord(cls, example, **kwargs):
        feature = {
            "input_ids": cls._int64_feature([int(x) for x in example.input_ids]),
            "segment_ids": cls._int64_feature([int(x) for x in example.segment_ids]),
            "attention_mask": cls._int64_feature([int(x) for x in example.attention_mask]),
            "pos_input_ids": cls._int64_feature([int(x) for x in example.pos_input_ids]),
            "pos_segment_ids": cls._int64_feature([int(x) for x in example.pos_segment_ids]),
            "pos_attention_mask": cls._int64_feature([int(x) for x in example.pos_attention_mask]),
        }
        return tf.train.Example(features=tf.train.Features(feature=feature))


class HardNegativeSimCSEDataset(AbstractDataset):
    """Dataset builder for hard negavtive SimCSE"""

    @classmethod
    def _filter(cls, dataset, max_sequence_length=512, **kwargs):
        dataset = dataset.filter(
            lambda a, b, c, e, f, g, h, i, j: tf.logical_and(
                tf.size(a) <= max_sequence_length,
                tf.logical_and(tf.size(e) <= max_sequence_length, tf.size(h) <= max_sequence_length),
            )
        )
        return dataset

    @classmethod
    def _bucket_padding(
        cls,
        dataset,
        batch_size=32,
        pad_id=0,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        drop_remainder=False,
        **kwargs
    ):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = cls._bucketing(
            dataset,
            element_length_func=lambda a, b, c, e, f, g, h, i, j: tf.maximum(
                tf.size(a), tf.maximum(tf.size(e), tf.size(h))
            ),
            padded_shapes=([None,], [None,], [None,], [None,], [None,], [None,], [None,], [None,], [None,]),
            padding_values=(pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id),
            batch_size=batch_size,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs,
        )
        # fmt: on
        return dataset

    @classmethod
    def _batch_padding(cls, dataset, batch_size=32, pad_id=0, drop_remainder=False, **kwargs):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([None,], [None,], [None,], [None,], [None,], [None,], [None,], [None,], [None,]),
            padding_values=(pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id),
            drop_remainder=drop_remainder,
        )
        # fmt: on
        return dataset

    @classmethod
    def _fixed_padding(cls, dataset, batch_size=32, pad_id=0, max_sequence_length=512, drop_remainder=False, **kwargs):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        maxlen = tf.constant(max_sequence_length, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([maxlen,], [maxlen,], [maxlen,], [maxlen,], [maxlen,], [maxlen,], [maxlen,], [maxlen,], [maxlen,]),
            padding_values=(pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id, pad_id),
            drop_remainder=drop_remainder,
        )
        # fmt: on
        return dataset

    @classmethod
    def _zip_dataset(cls, examples: List[UnsupervisedSimCSEExample], **kwargs):
        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset([e.input_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.segment_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.attention_mask for e in examples], dtype=tf.int32),
                _to_dataset([e.pos_input_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.pos_segment_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.pos_attention_mask for e in examples], dtype=tf.int32),
                _to_dataset([e.neg_input_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.neg_segment_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.neg_attention_mask for e in examples], dtype=tf.int32),
            )
        )
        return dataset

    @classmethod
    def _to_dict(cls, dataset, **kwargs):
        dataset = dataset.map(
            lambda a, b, c, e, f, g, h, i, j: (
                {
                    "input_ids": a,
                    "segment_ids": b,
                    "attention_mask": c,
                    "pos_input_ids": e,
                    "pos_segment_ids": f,
                    "pos_attention_mask": g,
                    "neg_input_ids": h,
                    "neg_segment_ids": i,
                    "neg_attention_mask": j,
                },
                None,
            ),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _parse_instances_to_examples(
        cls, instances, tokenizer: BertWordPieceTokenizer = None, vocab_file=None, **kwargs
    ):
        assert tokenizer or vocab_file, "`tokenizer` or `vocab_file` must be provided."
        if tokenizer is None:
            tokenizer = BertWordPieceTokenizer.from_file(
                vocab_file,
                lowercase=kwargs.get("do_lower_case", True),
            )
        examples = []
        for instance in instances:
            sequence = instance[kwargs.get("sequence_key", "sequence")]
            seq_encoding = tokenizer.encode(sequence)
            pos_sequence = instance[kwargs.get("pos_sequence_key", "pos_sequence")]
            pos_encoding = tokenizer.encode(pos_sequence)
            neg_sequence = instance[kwargs.get("neg_sequence_key", "neg_sequence")]
            neg_encoding = tokenizer.encode(neg_sequence)
            example = HardNegativeSimCSEExample(
                tokens=seq_encoding.tokens,
                input_ids=seq_encoding.ids,
                segment_ids=seq_encoding.type_ids,
                attention_mask=seq_encoding.attention_mask,
                pos_tokens=pos_encoding.tokens,
                pos_input_ids=pos_encoding.ids,
                pos_segment_ids=pos_encoding.type_ids,
                pos_attention_mask=pos_encoding.attention_mask,
                neg_tokens=neg_encoding.tokens,
                neg_input_ids=neg_encoding.ids,
                neg_segment_ids=neg_encoding.type_ids,
                neg_attention_mask=neg_encoding.attention_mask,
            )
            examples.append(example)
        return examples

    @classmethod
    def _parse_tfrecord(cls, dataset, **kwargs):
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "neg_input_ids": tf.io.VarLenFeature(tf.int64),
            "neg_segment_ids": tf.io.VarLenFeature(tf.int64),
            "neg_attention_mask": tf.io.VarLenFeature(tf.int64),
            "pos_input_ids": tf.io.VarLenFeature(tf.int64),
            "pos_segment_ids": tf.io.VarLenFeature(tf.int64),
            "pos_attention_mask": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["pos_input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["pos_segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["pos_attention_mask"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["neg_input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["neg_segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["neg_attention_mask"]), tf.int32),
            ),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _example_to_tfrecord(cls, example, **kwargs):
        feature = {
            "input_ids": cls._int64_feature([int(x) for x in example.input_ids]),
            "segment_ids": cls._int64_feature([int(x) for x in example.segment_ids]),
            "attention_mask": cls._int64_feature([int(x) for x in example.attention_mask]),
            "pos_input_ids": cls._int64_feature([int(x) for x in example.pos_input_ids]),
            "pos_segment_ids": cls._int64_feature([int(x) for x in example.pos_segment_ids]),
            "pos_attention_mask": cls._int64_feature([int(x) for x in example.pos_attention_mask]),
            "neg_input_ids": cls._int64_feature([int(x) for x in example.neg_input_ids]),
            "neg_segment_ids": cls._int64_feature([int(x) for x in example.neg_segment_ids]),
            "neg_attention_mask": cls._int64_feature([int(x) for x in example.neg_attention_mask]),
        }
        return tf.train.Example(features=tf.train.Features(feature=feature))



import collections
import logging

from . import utils


class ExactMatch:
    """Exact match"""

    def __call__(self, gold_spans, pred_spans, dim=1):
        if dim == 1:
            gold_spans_list, pred_spans_list = [gold_spans], [pred_spans]
        elif dim == 2:
            gold_spans_list, pred_spans_list = gold_spans, pred_spans
        else:
            raise ValueError("Invalid dim: " + str(dim))
        return self.call(gold_spans_list, pred_spans_list)

    def call(self, gold_spans_list, pred_spans_list):
        assert len(gold_spans_list) == len(pred_spans_list), "Length of inputs mismatch."
        matchs, total = 0, 0
        for gold_spans, pred_spans in zip(gold_spans_list, pred_spans_list):
            m, t = self._compute(gold_spans, pred_spans)
            matchs += m
            total += t
        return matchs * 1.0 / total

    def _compute(self, gold_spans, pred_spans):
        pred_spans = self._clip_pred_spans(gold_spans, pred_spans)
        num_matchs, num_total = 0, 0
        for idx in range(len(pred_spans)):
            pred = utils.normalize_span(pred_spans[idx])
            gold = utils.normalize_span(gold_spans[idx])
            num_matchs += pred == gold
        num_total += len(gold_spans)
        return num_matchs, num_total

    def _clip_pred_spans(self, gold_spans, pred_spans):
        if len(pred_spans) > len(gold_spans):
            # logging.warning("len(pred_spans) > len(gold_spans), clipping pred_spans...")
            pred_spans = pred_spans[: len(gold_spans)]
        return pred_spans


class F1ForSequence:
    """F1 for sequence tasks."""

    def __call__(self, gold_spans, pred_spans, dim=1, split_whitespace=False):
        if dim == 1:
            gold_spans_list, pred_spans_list = [gold_spans], [pred_spans]
        elif dim == 2:
            gold_spans_list, pred_spans_list = gold_spans, pred_spans
        else:
            raise ValueError("Invalid dim: " + str(dim))
        return self.call(gold_spans_list, pred_spans_list, split_whitespace=split_whitespace)

    def call(self, gold_spans_list, pred_spans_list, split_whitespace=False):
        assert len(gold_spans_list) == len(pred_spans_list), "Length of inputs mismatch."
        scores = []
        for gold_spans, pred_spans in zip(gold_spans_list, pred_spans_list):
            scores.extend(self._compute_f1(gold_spans, pred_spans, split_whitespace=split_whitespace))
        return sum(scores) / len(scores)

    def _compute_f1(self, gold_spans, pred_spans, split_whitespace=False):
        if len(pred_spans) > len(gold_spans):
            # logging.warning("len(pred_spans) > len(gold_spans), clipping pred_spans...")
            pred_spans = pred_spans[: len(gold_spans)]
        scores = []
        for idx in range(len(pred_spans)):
            gold = utils.normalize_span(gold_spans[idx])
            pred = utils.normalize_span(pred_spans[idx])
            scores.append(self._compute(gold, pred, split_whitespace=split_whitespace))
        return scores

    def _compute(self, pred, gold, split_whitespace=False):
        pred_toks = pred.split()
        gold_toks = gold.split()
        if not split_whitespace:
            pred_toks = pred_toks[0] if pred_toks else ""
            gold_toks = gold_toks[0] if gold_toks else ""
        common = collections.Counter(gold_toks) & collections.Counter(pred_toks)
        num_same = sum(common.values())
        if len(gold_toks) == 0 or len(pred_toks) == 0:
            # If either is no-answer, then F1 is 1 if they agree, 0 otherwise
            return int(gold_toks == pred_toks)
        if num_same == 0:
            return 0
        precision = 1.0 * num_same / len(pred_toks)
        recall = 1.0 * num_same / len(gold_toks)
        f1 = (2 * precision * recall) / (precision + recall)
        return f1



import logging
from typing import List

import numpy as np
import tensorflow as tf
from transformers_keras.common.metrics import ExactMatch, F1ForSequence

from .dataset import QuestionAnsweringDataset, QuestionAnsweringExample


class BaseMetricForQuestionAnswering(tf.keras.callbacks.Callback):
    """Base metric for qa."""

    @classmethod
    def from_jsonl_files(cls, input_files, vocab_file, limit=None, **kwargs):
        examples = QuestionAnsweringDataset.jsonl_to_examples(input_files, vocab_file=vocab_file, **kwargs)
        if limit is not None and limit > 0:
            examples = examples[:limit]
        return cls(examples, **kwargs)

    def __init__(self, examples, **kwargs):
        super().__init__()
        self.examples = examples
        self.dataset = QuestionAnsweringDataset.from_examples(self.examples, **kwargs)
        self.em = ExactMatch()

    def on_epoch_end(self, epoch, logs):
        outputs = self.model.predict(self.dataset)
        heads, tails = outputs[0], outputs[1]
        pred_answers, gold_answers = [], []
        for head, tail, example in zip(heads, tails, self.examples):
            head, tail = np.argmax(head), np.argmax(tail)
            pred_tokens = example.tokens[head : tail + 1]
            pred_text = "".join([str(x).lstrip("##") for x in pred_tokens])
            gold_text = "".join([str(x).lstrip("##") for x in example.tokens[example.start : example.end + 1]])
            pred_answers.append(pred_text)
            gold_answers.append(gold_text)
        self._compute_metric(gold_answers, pred_answers, epoch=epoch)

    def _compute_metric(self, gold_answers, pred_answers, epoch=0):
        raise NotImplementedError()


class ExactMatchForQuestionAnswering(BaseMetricForQuestionAnswering):
    """Exact Match metric for question answering."""

    def _compute_metric(self, gold_answers, pred_answers, epoch=0):
        acc = self.em(gold_answers, pred_answers, dim=1)
        tf.summary.scalar("EM", acc, step=epoch, description="EM Score")
        logging.info("No.%4d epoch EM: %.4f", epoch + 1, acc)


class F1ForQuestionAnswering(BaseMetricForQuestionAnswering):
    """F1 metric for question answering."""

    def __init__(self, examples: List[QuestionAnsweringExample], split_whitespace=False, **kwargs):
        super().__init__(examples, **kwargs)
        self.split_whitespace = split_whitespace
        self.f1 = F1ForSequence()

    def _compute_metric(self, gold_answers, pred_answers, epoch=0):
        score = self.f1(gold_answers, pred_answers, dim=1, split_whitespace=self.split_whitespace)
        tf.summary.scalar("F1", score, step=epoch, description="F1 Score")
        logging.info("No.%4d epoch F1: %.4f", epoch + 1, score)



import unittest

from transformers_keras.sentiment_analysis.dataset import AspectTermExtractionDataset


class DatasetTest(unittest.TestCase):
    """Dataset test"""

    def test_build_dataset(self):
        print("====from_jsonl_files\n")
        d = AspectTermExtractionDataset.from_jsonl_files(
            "testdata/ate.jsonl", vocab_file="testdata/vocab.bert.txt", batch_size=2
        )
        print(next(iter(d)))

        print("====jsonl_to_examples\n")
        examples = AspectTermExtractionDataset.jsonl_to_examples(
            "testdata/ate.jsonl", vocab_file="testdata/vocab.bert.txt"
        )
        for i in range(2):
            print(examples[i])

        print("====from_examples\n")
        d = AspectTermExtractionDataset.from_examples(examples, batch_size=2)
        print(next(iter(d)))

        print("====examples_to_tfrecord\n")
        AspectTermExtractionDataset.examples_to_tfrecord(examples, output_files=["testdata/ate.tfrecord"])

        print("====from_tfrecord_files\n")
        d = AspectTermExtractionDataset.from_tfrecord_files(input_files=["testdata/ate.tfrecord"])
        print(next(iter(d)))


if __name__ == "__main__":
    unittest.main()



import logging
import os
import re
from collections import namedtuple

import tensorflow as tf
from transformers_keras.common.abc_dataset import AbstractDataset
from transformers_keras.common.char_tokenizer import BertCharTokenizer

from .tokenizer import TokenClassificationLabelTokenizer

TokenClassificationExample = namedtuple(
    "TokenClassificationExample", ["tokens", "labels", "input_ids", "segment_ids", "attention_mask", "label_ids"]
)


class TokenClassificationDataset(AbstractDataset):
    """Dataset for token classification."""

    @classmethod
    def from_conll_files(
        cls,
        input_files,
        tokenizer=None,
        vocab_file=None,
        label_tokenizer=None,
        label_vocab_file=None,
        sep="\t",
        verbose=True,
        n=5,
        **kwargs
    ):
        examples = cls.conll_to_examples(
            input_files,
            tokenizer=tokenizer,
            vocab_file=vocab_file,
            label_tokenizer=label_tokenizer,
            label_vocab_file=label_vocab_file,
            sep=sep,
            **kwargs,
        )
        dataset = cls.from_examples(examples, verbose=verbose, n=n, **kwargs)
        return dataset

    @classmethod
    def conll_to_examples(
        cls,
        input_files,
        tokenizer=None,
        vocab_file=None,
        label_tokenizer=None,
        label_vocab_file=None,
        sep="\t",
        **kwargs
    ):
        instances = cls._read_conll(input_files, sep=sep, **kwargs)
        examples = cls._parse_instances_to_examples(
            instances,
            tokenizer=tokenizer,
            vocab_file=vocab_file,
            label_tokenizer=label_tokenizer,
            label_vocab_file=label_vocab_file,
            features_key="features",
            labels_ley="labels",
            **kwargs,
        )
        return examples

    @classmethod
    def _read_conll(cls, input_files, sep="\t", **kwargs):
        instances = []
        if isinstance(input_files, str):
            input_files = [input_files]
        for f in input_files:
            if not os.path.exists(f):
                logging.warning("File %s does not exist, skip.", f)
                continue
            features, labels = [], []
            with open(f, mode="rt", encoding="utf-8") as fin:
                for line in fin:
                    if not line:
                        break
                    line = line.strip()
                    if not line:
                        instances.append({"features": features, "labels": labels})
                        features, labels = [], []
                        continue
                    parts = re.split(sep, line)
                    if len(parts) < 2:
                        raise ValueError("Invalid file data")
                    features.append(str(parts[0].strip()))
                    labels.append(str(parts[1]).strip())
            if features and labels:
                instances.append({"features": features, "labels": labels})
        logging.info("Read %d instances from CoNLL files.", len(instances))
        return instances

    @classmethod
    def _filter(cls, dataset, max_sequence_length=512, **kwargs):
        dataset = dataset.filter(lambda a, b, c, y: tf.size(a) <= max_sequence_length)
        return dataset

    @classmethod
    def _bucket_padding(
        cls,
        dataset,
        batch_size=32,
        pad_id=0,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        drop_remainder=False,
        **kwargs
    ):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = cls._bucketing(
            dataset,
            element_length_func=lambda a, b, c, y: tf.size(a),
            padded_shapes=([None, ], [None, ], [None, ], [None, ]),
            padding_values=(pad_id, pad_id, pad_id, pad_id),
            batch_size=batch_size,
            pad_id=pad_id,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs,
        )
        # fmt: on
        return dataset

    @classmethod
    def _batch_padding(cls, dataset, batch_size=32, pad_id=0, drop_remainder=False, **kwargs):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([None, ], [None, ], [None, ], [None, ]),
            padding_values=(pad_id, pad_id, pad_id, pad_id),
            drop_remainder=drop_remainder,
        )
        # fmt: on
        return dataset

    @classmethod
    def _fixed_padding(cls, dataset, batch_size=32, pad_id=0, max_sequence_length=512, drop_remainder=False, **kwargs):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        maxlen = tf.constant(max_sequence_length, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([maxlen, ], [maxlen, ], [maxlen, ], [maxlen, ]),
            padding_values=(pad_id, pad_id, pad_id, pad_id),
            drop_remainder=drop_remainder,
        )
        # fmt: on
        return dataset

    @classmethod
    def _zip_dataset(cls, examples, **kwargs):
        """zip dataset"""

        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset([e.input_ids for e in examples]),
                _to_dataset([e.segment_ids for e in examples]),
                _to_dataset([e.attention_mask for e in examples]),
                _to_dataset([e.label_ids for e in examples]),
            )
        )
        return dataset

    @classmethod
    def _to_dict(cls, dataset, **kwargs):
        dataset = dataset.map(
            lambda a, b, c, y: ({"input_ids": a, "segment_ids": b, "attention_mask": c}, y),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _parse_instances_to_examples(
        cls,
        instances,
        tokenizer: BertCharTokenizer = None,
        vocab_file=None,
        label_tokenizer: TokenClassificationLabelTokenizer = None,
        label_vocab_file=None,
        **kwargs
    ):
        assert tokenizer or vocab_file, "`tokenizer` or `vocab_file` must be provided."
        if tokenizer is None:
            tokenizer = BertCharTokenizer.from_file(
                vocab_file,
                do_lower_case=kwargs.get("do_lower_case", True),
            )
        assert label_tokenizer or label_vocab_file, "`label_tokenizer` or `label_vocab_file` must be provided."
        if label_tokenizer is None:
            label_tokenizer = TokenClassificationLabelTokenizer.from_file(
                label_vocab_file, o_token=kwargs.get("o_token", "O")
            )
        # collect examples
        examples = []
        for instance in instances:
            tokens = instance[kwargs.get("features_key", "features")]
            input_ids = [tokenizer.token_to_id(token) for token in tokens]
            labels = instance[kwargs.get("labels_key", "labels")]
            label_ids = label_tokenizer.labels_to_ids(labels, add_cls=False, add_sep=False)
            example = TokenClassificationExample(
                tokens=tokens,
                labels=labels,
                input_ids=input_ids,
                segment_ids=[0] * len(input_ids),
                attention_mask=[1] * len(input_ids),
                label_ids=label_ids,
            )
            examples.append(example)
        return examples

    @classmethod
    def _parse_tfrecord(cls, dataset, **kwargs):
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "label_ids": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["label_ids"]), tf.int32),
            ),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _example_to_tfrecord(cls, example: TokenClassificationExample, **kwargs):
        feature = {
            "input_ids": cls._int64_feature([int(x) for x in example.input_ids]),
            "segment_ids": cls._int64_feature([int(x) for x in example.segment_ids]),
            "attention_mask": cls._int64_feature([int(x) for x in example.attention_mask]),
            "label_ids": cls._int64_feature([int(x) for x in example.label_ids]),
        }
        return tf.train.Example(features=tf.train.Features(feature=feature))



import abc
import json
import logging

import tensorflow as tf
from tensorflow.python.util.decorator_utils import classproperty

TF_VERSION = tf.__version__


class AbstractDataset(abc.ABC):
    """Abstract dataset builder"""

    @classproperty
    def AUTOTUNE(cls):
        try:
            autotune = tf.data.experimental.AUTOTUNE
            return autotune
        except Exception as e:
            try:
                autotune = tf.data.AUTOTUNE
                return autotune
            except Exception as e:
                logging.warning("Found AUTOTUNE exception: ", e)
        return None

    @classproperty
    def bucket_by_sequence_length(cls):
        try:
            fn = tf.data.experimental.bucket_by_sequence_length
            return fn
        except Exception as e:
            try:
                fn = tf.data.bucket_by_sequence_length
                return fn
            except Exception as e:
                logging.warning("Find bucket_by_sequence_length exception: ", e)
        return None

    @classmethod
    def from_examples(cls, examples, verbose=True, n=5, **kwargs):
        logging.info("Load %d examples in total.", len(examples))
        if verbose:
            cls._show_examples(examples, n=n, **kwargs)
        dataset = cls._zip_dataset(examples, **kwargs)
        dataset = cls._build_dataset(dataset, **kwargs)
        return dataset

    @classmethod
    def from_tfrecord_files(cls, input_files, **kwargs):
        dataset = cls._read_tfrecord(input_files, **kwargs)
        dataset = cls._build_dataset(dataset, **kwargs)
        return dataset

    @classmethod
    def from_jsonl_files(cls, input_files, tokenizer=None, vocab_file=None, **kwargs):
        instances = cls._read_jsonl(input_files, **kwargs)
        examples = cls._parse_instances_to_examples(instances, tokenizer=tokenizer, vocab_file=vocab_file, **kwargs)
        dataset = cls.from_examples(examples, **kwargs)
        return dataset

    @classmethod
    def jsonl_to_examples(cls, input_files, tokenizer=None, vocab_file=None, **kwargs):
        instances = cls._read_jsonl(input_files, **kwargs)
        examples = cls._parse_instances_to_examples(instances, tokenizer=tokenizer, vocab_file=vocab_file, **kwargs)
        return examples

    @classmethod
    def jsonl_to_tfrecord(cls, input_files, output_files, tokenizer=None, vocab_file=None, **kwargs):
        examples = cls.jsonl_to_examples(input_files, tokenizer=tokenizer, vocab_file=vocab_file, **kwargs)
        cls.examples_to_tfrecord(examples, output_files, **kwargs)

    @classmethod
    def examples_to_tfrecord(cls, examples, output_files, **kwargs):
        if isinstance(output_files, str):
            output_files = [output_files]
        writers = [tf.io.TFRecordWriter(f) for f in output_files]
        idx = 0
        for example in examples:
            tfrecord_example = cls._example_to_tfrecord(example)
            writers[idx].write(tfrecord_example.SerializeToString())
            idx += 1
            idx = idx % len(writers)
        for w in writers:
            w.close()
        logging.info("Finished to write %d examples to tfrecords.", len(examples))

    @classmethod
    def _example_to_tfrecord(cls, example, **kwargs):
        raise NotImplementedError()

    @classmethod
    def _zip_dataset(cls, examples, **kwargs):
        raise NotImplementedError()

    @classmethod
    def _build_dataset(
        cls,
        dataset,
        batch_size=64,
        max_sequence_length=512,
        padding="bucket",
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        repeat=None,
        shuffle=True,
        buffer_size=1000000,
        seed=None,
        reshuffle_each_iteration=True,
        pad_id=0,
        auto_shard_policy=None,
        drop_remainder=False,
        **kwargs,
    ):
        dataset = cls._filter(dataset, max_sequence_length=max_sequence_length, **kwargs)
        if shuffle is True:
            dataset = cls._shuffle(
                dataset,
                buffer_size=buffer_size,
                seed=seed,
                reshuffle_each_iteration=reshuffle_each_iteration,
                **kwargs,
            )
        dataset = cls._repeat(dataset, repeat=repeat, **kwargs)
        assert padding in ["bucket", "batch", "fixed"], "Invalid padding: " + str(padding)
        if padding == "bucket":
            dataset = cls._bucket_padding(
                dataset,
                pad_id=pad_id,
                batch_size=batch_size,
                bucket_boundaries=bucket_boundaries,
                bucket_batch_sizes=bucket_batch_sizes,
                drop_remainder=drop_remainder,
                **kwargs,
            )
        elif padding == "batch":
            dataset = cls._batch_padding(
                dataset,
                batch_size=batch_size,
                pad_id=pad_id,
                drop_remainder=drop_remainder,
                **kwargs,
            )
        elif padding == "fixed":
            dataset = cls._fixed_padding(
                dataset,
                batch_size=batch_size,
                pad_id=pad_id,
                max_sequence_length=max_sequence_length,
                drop_remainder=drop_remainder,
                **kwargs,
            )
        dataset = dataset.prefetch(cls.AUTOTUNE)
        dataset = cls._to_dict(dataset, **kwargs)
        dataset = cls._auto_shard(dataset, auto_shard_policy=auto_shard_policy, **kwargs)
        return dataset

    @classmethod
    def _filter(cls, dataset, max_sequence_length=512, **kwargs):
        raise NotImplementedError()

    @classmethod
    def _shuffle(cls, dataset, buffer_size, seed=None, reshuffle_each_iteration=True, **kwargs):
        dataset = dataset.shuffle(buffer_size=buffer_size, seed=seed, reshuffle_each_iteration=reshuffle_each_iteration)
        return dataset

    @classmethod
    def _repeat(cls, dataset, repeat=None, **kwargs):
        if repeat is None:
            return dataset
        dataset = dataset.repeat(repeat)
        return dataset

    @classmethod
    def _bucket_padding(
        cls,
        dataset,
        batch_size=32,
        pad_id=0,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        drop_remainder=False,
        **kwargs,
    ):
        raise NotImplementedError()

    @classmethod
    def _batch_padding(cls, dataset, batch_size=32, pad_id=0, drop_remainder=False, **kwargs):
        raise NotImplementedError()

    @classmethod
    def _fixed_padding(cls, dataset, batch_size=32, pad_id=0, max_sequence_length=512, drop_remainder=False, **kwargs):
        raise NotImplementedError()

    @classmethod
    def _to_dict(cls, dataset, **kwargs):
        raise NotImplementedError()

    @classmethod
    def _bucketing(
        cls,
        dataset,
        element_length_func,
        padded_shapes,
        padding_values,
        batch_size=64,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        drop_remainder=False,
        **kwargs,
    ):
        if bucket_batch_sizes is None:
            bucket_batch_sizes = [batch_size] * (len(bucket_boundaries) + 1)
        assert (
            len(bucket_batch_sizes) == len(bucket_boundaries) + 1
        ), "len(bucket_batch_sizes) should equals len(bucket_doundaries) + 1"

        # fmt: off
        dataset = dataset.apply(cls.bucket_by_sequence_length(
            element_length_func=element_length_func,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            padded_shapes=padded_shapes,
            padding_values=padding_values,
            drop_remainder=drop_remainder,
        )).prefetch(cls.AUTOTUNE)
        # fmt: on
        return dataset

    @classmethod
    def _read_jsonl(cls, input_files, **kwargs):
        if isinstance(input_files, str):
            input_files = [input_files]
        instances = []
        for f in input_files:
            with open(f, mode="rt", encoding="utf-8") as fin:
                for line in fin:
                    line = line.strip()
                    if not line:
                        continue
                    instance = json.loads(line)
                    instances.append(instance)
        logging.info("Collected %d instances in total.", len(instances))
        return instances

    @classmethod
    def _parse_instances_to_examples(cls, instances, tokenizer=None, vocab_file=None, **kwargs):
        raise NotImplementedError()

    @classmethod
    def _read_tfrecord(cls, input_files, **kwargs):
        if isinstance(input_files, str):
            input_files = [input_files]
        if len(input_files) == 1:
            dataset = tf.data.TFRecordDataset(input_files)
        else:
            dataset = tf.data.Dataset.from_tensor_slices(input_files)
            dataset = dataset.interleave(
                lambda x: tf.data.TFRecordDataset(x),
                cycle_length=len(input_files),
                num_parallel_calls=cls.AUTOTUNE,
            )
        dataset = cls._parse_tfrecord(dataset, **kwargs)
        return dataset

    @classmethod
    def _parse_tfrecord(cls, dataset, **kwargs):
        raise NotImplementedError()

    @classmethod
    def _auto_shard(cls, dataset, auto_shard_policy=None, **kwargs):
        if auto_shard_policy is not None:
            options = tf.data.Options()
            # options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
            options.experimental_distribute.auto_shard_policy = auto_shard_policy
            dataset = dataset.with_options(options)
        return dataset

    @classmethod
    def _show_examples(cls, examples, n=5, **kwargs):
        n = min(n, len(examples))
        logging.info("Showing %d examples.", n)
        for i in range(n):
            logging.info("NO.%d example: %s", i, examples[i])

    @classmethod
    def _int64_feature(cls, values):
        return tf.train.Feature(int64_list=tf.train.Int64List(value=values))



import logging
from collections import namedtuple
from typing import List

import tensorflow as tf
from transformers_keras.common.abc_dataset import AbstractDataset
from transformers_keras.common.char_tokenizer import BertCharTokenizer
from transformers_keras.question_answering.dataset import QuestionAnsweringXDataset

AspectTermExtractionExample = namedtuple(
    "AspectTermExtractionExample",
    ["tokens", "input_ids", "segment_ids", "attention_mask", "start_ids", "end_ids"],
)


class AspectTermExtractionDataset(AbstractDataset):
    """Build dataset for aspect term extraction models."""

    @classmethod
    def _filter(cls, dataset, max_sequence_length=512, **kwargs):
        dataset = dataset.filter(
            lambda a, b, c, x, y: tf.size(a) <= max_sequence_length,
        )
        return dataset

    @classmethod
    def _bucket_padding(
        cls,
        dataset,
        batch_size=32,
        pad_id=0,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        drop_remainder=False,
        **kwargs,
    ):
        # fmt: off
        dataset = cls._bucketing(
            dataset,
            element_length_func=lambda a, b, c, x, y: tf.size(a),
            padded_shapes=([None, ], [None, ], [None, ], [None, ], [None, ]),
            padding_values=(pad_id, pad_id, pad_id, pad_id, pad_id),
            batch_size=batch_size,
            pad_id=pad_id,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs,
        )
        # fmt: on
        return dataset

    @classmethod
    def _batch_padding(cls, dataset, batch_size=32, pad_id=0, drop_remainder=False, **kwargs):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([None, ], [None, ], [None, ], [None, ], [None, ]),
            padding_values=(pad_id, pad_id, pad_id, pad_id, pad_id),
            drop_remainder=drop_remainder,
        )
        # fmt: on
        return dataset

    @classmethod
    def _fixed_padding(
        cls,
        dataset,
        batch_size=32,
        pad_id=0,
        max_sequence_length=512,
        drop_remainder=False,
        **kwargs,
    ):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        maxlen = tf.constant(max_sequence_length, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([maxlen, ], [maxlen, ], [maxlen, ], [maxlen, ], [maxlen, ]),
            padding_values=(pad_id, pad_id, pad_id, pad_id, pad_id),
            drop_remainder=drop_remainder,
        )
        # fmt: on
        return dataset

    @classmethod
    def _to_dict(cls, dataset, **kwargs):
        dataset = dataset.map(
            lambda a, b, c, x, y: ({"input_ids": a, "segment_ids": b, "attention_mask": c}, {"start": x, "end": y}),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _zip_dataset(cls, examples: List[AspectTermExtractionExample], **kwargs):
        """Zip datasets to one dataset."""

        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset(x=[e.input_ids for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.segment_ids for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.attention_mask for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.start_ids for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.end_ids for e in examples], dtype=tf.int32),
            )
        )
        return dataset

    @classmethod
    def _parse_instances_to_examples(
        cls, instances, tokenizer: BertCharTokenizer = None, vocab_file=None, question="文中提到了哪些层面？", **kwargs
    ):
        assert tokenizer or vocab_file, "`tokenizer` or `vocab_file` must be provided."
        if tokenizer is None:
            tokenizer = BertCharTokenizer.from_file(
                vocab_file,
                do_lower_case=kwargs.get("do_lower_case", True),
            )
        examples = []
        for instance in instances:
            sequence = instance[kwargs.get("sequence_key", "sequence")]
            sequence_encoding = tokenizer.encode(sequence, add_cls=True, add_sep=True)
            question_encoding = tokenizer.encode(question, add_cls=False, add_sep=True)
            aspect_spans = instance[kwargs.get("aspect_spans_key", "aspect_spans")]
            input_length = len(sequence_encoding.ids) + len(question_encoding.ids)
            start_ids, end_ids = [0] * input_length, [0] * input_length
            tokens = sequence_encoding.tokens + question_encoding.tokens
            for span in aspect_spans:
                # start + 1: [CLS] offset
                start, end = span[0] + 1, span[1]
                assert "".join(tokens[start : end + 1]).lower() == str(sequence[span[0] : span[1]]).lower()
                start_ids[start] = 1
                end_ids[end] = 1
            example = AspectTermExtractionExample(
                tokens=tokens,
                input_ids=sequence_encoding.ids + question_encoding.ids,
                segment_ids=[0] * len(sequence_encoding.ids) + [1] * len(question_encoding.ids),
                attention_mask=[1] * (len(sequence_encoding.ids) + len(question_encoding.ids)),
                start_ids=start_ids,
                end_ids=end_ids,
            )
            examples.append(example)
        logging.info("Collected %d examples in total.", len(examples))
        return examples

    @classmethod
    def _parse_tfrecord(cls, dataset, **kwargs):
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "start": tf.io.VarLenFeature(tf.int64),
            "end": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["start"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["end"]), tf.int32),
            ),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _example_to_tfrecord(cls, example: AspectTermExtractionExample, **kwargs):
        feature = {
            "input_ids": cls._int64_feature([int(x) for x in example.input_ids]),
            "segment_ids": cls._int64_feature([int(x) for x in example.segment_ids]),
            "attention_mask": cls._int64_feature([int(x) for x in example.attention_mask]),
            "start": cls._int64_feature([int(x) for x in example.start_ids]),
            "end": cls._int64_feature([int(x) for x in example.end_ids]),
        }
        return tf.train.Example(features=tf.train.Features(feature=feature))


class OpinionTermExtractionAndClassificationDataset(QuestionAnsweringXDataset):
    """Dataset builder for Opinion Term Extraction and Classification"""

    pass



import logging
import re
from collections import namedtuple
from typing import List

import tensorflow as tf
from tokenizers import BertWordPieceTokenizer
from transformers_keras.common.abc_dataset import AbstractDataset

QuestionAnsweringExample = namedtuple(
    "QuestionAnsweringExample", ["tokens", "input_ids", "segment_ids", "attention_mask", "start", "end"]
)
QuestionAnsweringXExample = namedtuple(
    "QuestionAnsweringXExample",
    ["tokens", "input_ids", "segment_ids", "attention_mask", "start", "end", "class_id"],
)


class QuestionAnsweringDataset(AbstractDataset):
    """Dataset builder for question answering models."""

    @classmethod
    def _zip_dataset(cls, examples: List[QuestionAnsweringExample], **kwargs):
        """Zip datasets to one dataset."""

        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset(x=[e.input_ids for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.segment_ids for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.attention_mask for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.start for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.end for e in examples], dtype=tf.int32),
            )
        )
        return dataset

    @classmethod
    def _filter(cls, dataset, max_sequence_length=512, **kwargs):
        dataset = dataset.filter(
            lambda a, b, c, x, y: tf.size(a) <= max_sequence_length,
        )
        return dataset

    @classmethod
    def _bucket_padding(
        cls,
        dataset,
        batch_size=32,
        pad_id=0,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        drop_remainder=False,
        **kwargs,
    ):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = cls._bucketing(
            dataset,
            element_length_func=lambda a, b, c, x, y: tf.size(a),
            padded_shapes=([None, ], [None, ], [None, ], [], []),
            padding_values=(pad_id, pad_id, pad_id, None, None),
            batch_size=batch_size,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs,
        )
        # fmt: on
        return dataset

    @classmethod
    def _batch_padding(cls, dataset, batch_size=32, pad_id=0, drop_remainder=False, **kwargs):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([None, ], [None, ], [None, ], [], []),
            padding_values=(pad_id, pad_id, pad_id, None, None),
            drop_remainder=drop_remainder,
        ).prefetch(cls.AUTOTUNE)
        # fmt: on
        return dataset

    @classmethod
    def _fixed_padding(cls, dataset, batch_size=32, pad_id=0, max_sequence_length=512, drop_remainder=False, **kwargs):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        maxlen = tf.constant(max_sequence_length, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([maxlen, ], [maxlen, ], [maxlen, ], [], []),
            padding_values=(pad_id, pad_id, pad_id, None, None),
            drop_remainder=drop_remainder,
        ).prefetch(cls.AUTOTUNE)
        # fmt: on
        return dataset

    @classmethod
    def _to_dict(cls, dataset, **kwargs):
        dataset = dataset.map(
            lambda a, b, c, x, y: ({"input_ids": a, "segment_ids": b, "attention_mask": c}, {"head": x, "tail": y}),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _parse_tfrecord(cls, dataset, **kwargs):
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "start": tf.io.VarLenFeature(tf.int64),
            "end": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.squeeze(tf.sparse.to_dense(x["start"])), tf.int32),
                tf.cast(tf.squeeze(tf.sparse.to_dense(x["end"])), tf.int32),
            ),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _example_to_tfrecord(cls, example: QuestionAnsweringExample, **kwargs):
        feature = {
            "input_ids": cls._int64_feature([int(x) for x in example.input_ids]),
            "segment_ids": cls._int64_feature([int(x) for x in example.segment_ids]),
            "attention_mask": cls._int64_feature([int(x) for x in example.attention_mask]),
            "start": cls._int64_feature([int(example.start)]),
            "end": cls._int64_feature([int(example.end)]),
        }
        return tf.train.Example(features=tf.train.Features(feature=feature))

    @classmethod
    def _parse_instances_to_examples(
        cls, instances, tokenizer: BertWordPieceTokenizer = None, vocab_file=None, **kwargs
    ):
        assert tokenizer or vocab_file, "`tokenizer` or `vocab_file` must be provided."
        if tokenizer is None:
            tokenizer = BertWordPieceTokenizer.from_file(
                vocab_file,
                lowercase=kwargs.get("do_lower_case", True),
            )
        examples = []
        for instance in instances:
            example = cls._instance_to_example(instance, tokenizer, **kwargs)
            if not example:
                continue
            examples.append(example)
        logging.info("Collected %d examples in total.", len(examples))
        return examples

    @classmethod
    def _instance_to_example(cls, instance, tokenizer: BertWordPieceTokenizer, **kwargs):
        context = instance[kwargs.get("context_key", "context")]
        question = instance[kwargs.get("question_key", "question")]
        answer = instance[kwargs.get("answer_key", "answer")]
        start_char_idx, end_char_idx = cls._find_answer_span(context, answer)
        if end_char_idx <= start_char_idx:
            return None
        # Mark the character indexes in context that are in answer
        is_char_in_ans = [0] * len(context)
        for idx in range(start_char_idx, end_char_idx):
            is_char_in_ans[idx] = 1
        context_encoding = tokenizer.encode(context)
        # Find tokens that were created from answer characters
        ans_token_idx = []
        for idx, (start_char_idx, end_char_idx) in enumerate(context_encoding.offsets):
            if sum(is_char_in_ans[start_char_idx:end_char_idx]) > 0:
                ans_token_idx.append(idx)
        if not ans_token_idx:
            return None
        start_token_idx, end_token_idx = ans_token_idx[0], ans_token_idx[-1]
        question_encoding = tokenizer.encode(question)
        input_ids = context_encoding.ids + question_encoding.ids[1:]
        segment_ids = [0] * len(context_encoding.type_ids) + [1] * len(question_encoding.type_ids[1:])
        attention_mask = [1] * len(context_encoding.attention_mask + question_encoding.attention_mask[1:])
        assert len(input_ids) == len(segment_ids), "input_ids length:{} VS segment_ids length: {}".format(
            len(input_ids), len(segment_ids)
        )
        assert len(input_ids) == len(attention_mask), "input_ids length:{} VS attention_mask length: {}".format(
            len(input_ids), len(attention_mask)
        )

        example = QuestionAnsweringExample(
            tokens=context_encoding.tokens + question_encoding.tokens[1:],
            input_ids=input_ids,
            segment_ids=segment_ids,
            attention_mask=attention_mask,
            start=start_token_idx,
            end=end_token_idx,
        )
        return example

    @classmethod
    def _find_answer_span(cls, context, answer):
        for m in re.finditer(re.escape(answer), context, re.IGNORECASE):
            start, end = m.span()
            return start, end
        return 0, 0


class QuestionAnsweringXDataset(AbstractDataset):
    """Dataset builder for question answering models."""

    @classmethod
    def _filter(cls, dataset, max_sequence_length=512, **kwargs):
        dataset = dataset.filter(
            lambda a, b, c, x, y, z: tf.size(a) <= max_sequence_length,
        )
        return dataset

    @classmethod
    def _bucket_padding(
        cls,
        dataset,
        batch_size=32,
        pad_id=0,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        drop_remainder=False,
        **kwargs,
    ):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = cls._bucketing(
            dataset,
            element_length_func=lambda a, b, c, x, y, z: tf.size(a),
            padded_shapes=([None, ], [None, ], [None, ], [], [], []),
            padding_values=(pad_id, pad_id, pad_id, None, None, None),
            batch_size=batch_size,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs,
        )
        # fmt: on
        return dataset

    @classmethod
    def _batch_padding(cls, dataset, batch_size=32, pad_id=0, drop_remainder=False, **kwargs):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([None, ], [None, ], [None, ], [], [], []),
            padding_values=(pad_id, pad_id, pad_id, None, None, None),
            drop_remainder=drop_remainder,
        )
        # fmt: on
        return dataset

    @classmethod
    def _fixed_padding(cls, dataset, batch_size=32, pad_id=0, max_sequence_length=512, drop_remainder=False, **kwargs):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        maxlen = tf.constant(max_sequence_length, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([maxlen, ], [maxlen, ], [maxlen, ], [], [], []),
            padding_values=(pad_id, pad_id, pad_id, None, None, None),
            drop_remainder=drop_remainder,
        )
        # fmt: on
        return dataset

    @classmethod
    def _zip_dataset(cls, examples: List[QuestionAnsweringXExample], **kwargs):
        """Zip datasets to one dataset."""

        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset(x=[e.input_ids for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.segment_ids for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.attention_mask for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.start for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.end for e in examples], dtype=tf.int32),
                _to_dataset(x=[e.class_id for e in examples], dtype=tf.int32),
            )
        )
        return dataset

    @classmethod
    def _to_dict(cls, dataset, **kwargs):
        dataset = dataset.map(
            lambda a, b, c, x, y, z: (
                {"input_ids": a, "segment_ids": b, "attention_mask": c},
                {"head": x, "tail": y, "class": z},
            ),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _parse_tfrecord(cls, dataset, **kwargs):
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "start": tf.io.VarLenFeature(tf.int64),
            "end": tf.io.VarLenFeature(tf.int64),
            "class_id": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.squeeze(tf.sparse.to_dense(x["start"])), tf.int32),
                tf.cast(tf.squeeze(tf.sparse.to_dense(x["end"])), tf.int32),
                tf.cast(tf.squeeze(tf.sparse.to_dense(x["class_id"])), tf.int32),
            ),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _example_to_tfrecord(cls, example: QuestionAnsweringXExample, **kwargs):
        feature = {
            "input_ids": cls._int64_feature([int(x) for x in example.input_ids]),
            "segment_ids": cls._int64_feature([int(x) for x in example.segment_ids]),
            "attention_mask": cls._int64_feature([int(x) for x in example.attention_mask]),
            "start": cls._int64_feature([int(example.start)]),
            "end": cls._int64_feature([int(example.end)]),
            "class_id": cls._int64_feature([int(example.class_id)]),
        }
        return tf.train.Example(features=tf.train.Features(feature=feature))

    @classmethod
    def _parse_instances_to_examples(
        cls, instances, tokenizer: BertWordPieceTokenizer = None, vocab_file=None, **kwargs
    ):
        assert tokenizer or vocab_file, "`tokenizer` or `vocab_file` must be provided."
        if tokenizer is None:
            tokenizer = BertWordPieceTokenizer.from_file(
                vocab_file,
                lowercase=kwargs.get("do_lower_case", True),
            )
        examples = []
        for instance in instances:
            example = cls._instance_to_example(instance, tokenizer, **kwargs)
            if not example:
                continue
            examples.append(example)
        logging.info("Collected %d examples in total.", len(examples))
        return examples

    @classmethod
    def _instance_to_example(cls, instance, tokenizer: BertWordPieceTokenizer, **kwargs):
        context = instance[kwargs.get("context_key", "context")]
        question = instance[kwargs.get("question_key", "question")]
        answer = instance[kwargs.get("answer_key", "answer")]
        class_id = instance[kwargs.get("class_key", "class")]

        context_encoding = tokenizer.encode(context)
        question_encoding = tokenizer.encode(question)

        start_char_idx, end_char_idx = cls._find_answer_span(context, answer)
        # invalid answer index
        if end_char_idx <= start_char_idx:
            return cls._invalid_answer_example(instance, context_encoding, question_encoding, **kwargs)

        # Mark the character indexes in context that are in answer
        is_char_in_ans = [0] * len(context)
        for idx in range(start_char_idx, end_char_idx):
            is_char_in_ans[idx] = 1
        # Find tokens that were created from answer characters
        ans_token_idx = []
        for idx, (start_char_idx, end_char_idx) in enumerate(context_encoding.offsets):
            if sum(is_char_in_ans[start_char_idx:end_char_idx]) > 0:
                ans_token_idx.append(idx)
        # set token index of start & end
        start_token_idx = ans_token_idx[0] if ans_token_idx else 0
        end_token_idx = ans_token_idx[-1] if ans_token_idx else 0
        # build example
        tokens = context_encoding.tokens + question_encoding.tokens[1:]
        input_ids = context_encoding.ids + question_encoding.ids[1:]
        segment_ids = [0] * len(context_encoding.type_ids) + [1] * len(context_encoding.type_ids[1:])
        attention_mask = [1] * len(context_encoding.attention_mask + question_encoding.attention_mask[1:])
        return QuestionAnsweringXExample(
            tokens=tokens,
            input_ids=input_ids,
            segment_ids=segment_ids,
            attention_mask=attention_mask,
            start=start_token_idx,
            end=end_token_idx,
            class_id=class_id,
        )

    @classmethod
    def _find_answer_span(cls, context, answer):
        for m in re.finditer(re.escape(answer), context, re.IGNORECASE):
            start, end = m.span()
            return start, end
        return 0, 0



from collections import namedtuple

import tensorflow as tf
from tokenizers import BertWordPieceTokenizer
from transformers_keras.common.abc_dataset import AbstractDataset

SequenceClassificationExample = namedtuple(
    "SequenceClassificationExample", ["tokens", "input_ids", "segment_ids", "attention_mask", "label"]
)


class SequenceClassificationDataset(AbstractDataset):
    """Dataset builder for sequence classification."""

    @classmethod
    def _filter(cls, dataset, max_sequence_length=512, **kwargs):
        dataset = dataset.filter(
            lambda a, b, c, y: tf.size(a) <= max_sequence_length,
        )
        return dataset

    @classmethod
    def _bucket_padding(
        cls,
        dataset,
        batch_size=32,
        pad_id=0,
        bucket_boundaries=[50, 100, 150, 200, 250, 300, 350, 400, 450, 500],
        bucket_batch_sizes=None,
        drop_remainder=False,
        **kwargs
    ):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = cls._bucketing(
            dataset,
            element_length_func=lambda a, b, c, y: tf.size(a),
            padded_shapes=([None, ], [None, ], [None, ], []),
            padding_values=(pad_id, pad_id, pad_id, None),
            batch_size=batch_size,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs
        )
        # fmt: on
        return dataset

    @classmethod
    def _batch_padding(cls, dataset, batch_size=32, pad_id=0, drop_remainder=False, **kwargs):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([None, ], [None, ], [None, ], []),
            padding_values=(pad_id, pad_id, pad_id, None),
            drop_remainder=drop_remainder,
        )
        # fmt: on
        return dataset

    @classmethod
    def _fixed_padding(cls, dataset, batch_size=32, pad_id=0, max_sequence_length=512, drop_remainder=False, **kwargs):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        maxlen = tf.constant(max_sequence_length, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([maxlen, ], [maxlen, ], [maxlen, ], []),
            padding_values=(pad_id, pad_id, pad_id, None),
            drop_remainder=drop_remainder,
        )
        # fmt: on
        return dataset

    @classmethod
    def _zip_dataset(cls, examples, **kwargs):
        """zip dataset"""

        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset([e.input_ids for e in examples]),
                _to_dataset([e.segment_ids for e in examples]),
                _to_dataset([e.attention_mask for e in examples]),
                _to_dataset([e.label for e in examples]),
            )
        )
        return dataset

    @classmethod
    def _to_dict(cls, dataset, **kwargs):
        dataset = dataset.map(
            lambda a, b, c, y: ({"input_ids": a, "segment_ids": b, "attention_mask": c}, y),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _parse_tfrecord(cls, dataset, **kwargs):
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "label": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.squeeze(tf.sparse.to_dense(x["label"])), tf.int32),
            ),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _example_to_tfrecord(cls, example: SequenceClassificationExample, **kwargs):
        feature = {
            "input_ids": cls._int64_feature([int(x) for x in example.input_ids]),
            "segment_ids": cls._int64_feature([int(x) for x in example.segment_ids]),
            "attention_mask": cls._int64_feature([int(x) for x in example.attention_mask]),
            "label": cls._int64_feature([int(example.label)]),
        }
        return tf.train.Example(features=tf.train.Features(feature=feature))

    @classmethod
    def _parse_instances_to_examples(
        cls, instances, tokenizer: BertWordPieceTokenizer = None, vocab_file=None, **kwargs
    ):
        assert tokenizer or vocab_file, "`tokenizer` or `vocab_file` must be provided."
        if tokenizer is None:
            tokenizer = BertWordPieceTokenizer.from_file(
                vocab_file,
                lowercase=kwargs.get("do_lower_case", True),
            )
        examples = []
        for instance in instances:
            sequence = instance[kwargs.get("sequence_key", "sequence")]
            label = instance[kwargs.get("label_key", "label")]
            encoding = tokenizer.encode(sequence)
            examples.append(
                SequenceClassificationExample(
                    tokens=encoding.tokens,
                    input_ids=encoding.ids,
                    segment_ids=encoding.type_ids,
                    attention_mask=encoding.attention_mask,
                    label=int(label),
                )
            )
        return examples



import json
import logging

import tensorflow as tf

from .abstract_adapter import AbstractAdapter, zip_weights


class BertAdapter(AbstractAdapter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def adapte_config(self, config_file, **kwargs):
        with open(config_file, mode="rt", encoding="utf8") as fin:
            config = json.load(fin)

        model_config = {
            "vocab_size": config["vocab_size"],
            "activation": config["hidden_act"],
            "max_positions": config["max_position_embeddings"],
            "hidden_size": config["hidden_size"],
            "type_vocab_size": config["type_vocab_size"],
            "intermediate_size": config["intermediate_size"],
            "hidden_dropout_rate": config["hidden_dropout_prob"],
            "attention_dropout_rate": config["attention_probs_dropout_prob"],
            "initializer_range": config["initializer_range"],
            "num_layers": config["num_hidden_layers"],
            "num_attention_heads": config["num_attention_heads"],
        }
        return model_config

    def adapte_weights(self, model, ckpt, model_config, use_functional_api=True, **kwargs):
        ckpt_weight_names = [x[0] for x in tf.train.list_variables(ckpt)]
        self_weight_names = set([x.name for x in model.trainable_weights])
        weight_mapping = {}
        bert_weight_mapping = self._adapte_bert_weights(model, ckpt, use_functional_api=use_functional_api, **kwargs)
        weight_mapping.update(bert_weight_mapping)

        mlm_weight_mapping = self._adapte_mlm_weights(model, ckpt, use_functional_api=use_functional_api, **kwargs)
        weight_mapping.update(mlm_weight_mapping)

        nsp_weight_mapping = self._adapte_nsp_weights(model, ckpt, use_functional_api=use_functional_api, **kwargs)
        weight_mapping.update(nsp_weight_mapping)

        check_weights = kwargs.get("check_weights", False)
        if check_weights:
            self._check_weights(weight_mapping, model, ckpt)

    # apdate BERT backbone's weights
    def _adapte_bert_weights(self, model, ckpt, use_functional_api=True, **kwargs):
        model_prefix = model.bert_model.name if use_functional_api else model.name + "/" + model.bert_model.name
        logging.info("Adapting bert weights, using model weight prefix: %s", model_prefix)
        ckpt_prefix = kwargs.get("ckpt_bert_prefix", "bert")
        logging.info("Adapting bert weights, using  ckpt weight prefix: %s", ckpt_prefix)
        ckpt_weight_names = [x for (x, _) in tf.train.list_variables(ckpt) if str(x).startswith(ckpt_prefix)]
        self_weight_names = set([x.name for x in model.trainable_weights if str(x.name).startswith(model_prefix)])
        mapping = {}
        # collect weights mapping
        for w in ckpt_weight_names:
            if any(x in w for x in ["embeddings", "pooler", "encoder"]):
                mw = model_prefix + w.lstrip(ckpt_prefix) + ":0"
                if mw not in self_weight_names:
                    logging.warning("weight: %s not in model weights", mw)
                    continue
                mapping[mw] = w
        # skip weights
        if self.skip_token_embedding:
            self._skip_weight(mapping, model_prefix + "/embeddings/word_embeddings:0")
        if self.skip_position_embedding:
            self._skip_weight(mapping, model_prefix + "/embeddings/position_embeddings:0")
        if self.skip_segment_embedding:
            self._skip_weight(mapping, model_prefix + "/embeddings/token_type_embeddings:0")
        if self.skip_embedding_layernorm:
            self._skip_weight(mapping, model_prefix + "/embeddings/LayerNorm/gamma:0")
            self._skip_weight(mapping, model_prefix + "/embeddings/LayerNorm/beta:0")
        if self.skip_pooler:
            self._skip_weight(mapping, model_prefix + "/pooler/dense/kernel:0")
            self._skip_weight(mapping, model_prefix + "/pooler/dense/bias:0")

        # zip weight names and values
        zipped_weights = zip_weights(model, ckpt, mapping, self_weight_names, **kwargs)
        # set values to weights
        tf.keras.backend.batch_set_value(zipped_weights)

        return mapping

    # adapte MLM's weights
    def _adapte_mlm_weights(self, model, ckpt, use_functional_api=True, **kwargs):
        with_mlm = kwargs.get("with_mlm", False)
        if not with_mlm:
            logging.info("Skipping to adapte weights for MLM due to option `with_mlm` set to `False`")
            return {}
        self_mlm_prefix = "cls/predictions" if use_functional_api else model.name + "/cls/predictions"
        ckpt_mlm_prefix = kwargs.get("ckpt_mlm_prefix", "cls/predictions")
        logging.info("Adapting MLM weights, using model weight prefix: %s", self_mlm_prefix)
        logging.info("Adapting MLM weights, using  ckpt weight prefix: %s", ckpt_mlm_prefix)
        ckpt_weight_names = [x for (x, _) in tf.train.list_variables(ckpt) if str(x).startswith(ckpt_mlm_prefix)]
        self_weight_names = set([x.name for x in model.trainable_weights if str(x.name).startswith(self_mlm_prefix)])
        mapping = {}
        for w in ckpt_weight_names:
            if any(k in w for k in ["key", "value", "query", "prev"]):
                continue
            if str(w).startswith(ckpt_mlm_prefix):
                mw = self_mlm_prefix + w[len(ckpt_mlm_prefix) :] + ":0"
                if mw not in self_weight_names:
                    logging.warning("weight: %s not in model weights", mw)
                    continue
                mapping[mw] = w

        # zip weight names and values
        zipped_weights = zip_weights(model, ckpt, mapping, self_weight_names, **kwargs)
        # set values to weights
        tf.keras.backend.batch_set_value(zipped_weights)

        return mapping

    # adapte NSP's weights
    def _adapte_nsp_weights(self, model, ckpt, use_functional_api=True, **kwargs):
        with_nsp = kwargs.get("with_nsp", False)
        if not with_nsp:
            logging.info("Skipping to adapte weights for NSP due to option `with_nsp` set to `False`")
            return {}
        self_nsp_prefix = "cls/seq_relationship" if use_functional_api else model.name + "/cls/seq_relationship"
        ckpt_nsp_prefix = kwargs.get("ckpt_nsp_prefix", "cls/seq_relationship")
        logging.info("Adapting NSP weights, using model weight prefix: %s", self_nsp_prefix)
        logging.info("Adapting NSP weights, using  ckpt weight prefix: %s", ckpt_nsp_prefix)
        ckpt_weight_names = [x for (x, _) in tf.train.list_variables(ckpt) if str(x).startswith(ckpt_nsp_prefix)]
        self_weight_names = set([x.name for x in model.trainable_weights if str(x.name).startswith(self_nsp_prefix)])
        mapping = {}
        for w in ckpt_weight_names:
            if str(w).startswith(ckpt_nsp_prefix):
                mw = self_nsp_prefix + w.lstrip(ckpt_nsp_prefix) + ":0"
                if mw not in self_weight_names:
                    logging.warning("weight: %s not in model weights", mw)
                    continue
                mapping[mw] = w
        # zip weight names and values
        zipped_weights = zip_weights(model, ckpt, mapping, self_weight_names, **kwargs)
        # set values to weights
        tf.keras.backend.batch_set_value(zipped_weights)

        return mapping



from collections import namedtuple

TokenizerEncoding = namedtuple("TokenizerEncoding", ["text", "tokens", "ids", "type_ids", "attention_mask", "offsets"])


class BertCharTokenizer:
    """Bert char level tokenizer for token classification tasks."""

    def __init__(
        self,
        token2id,
        do_lower_case=True,
        unk_token="[UNK]",
        pad_token="[PAD]",
        cls_token="[CLS]",
        sep_token="[SEP]",
        mask_token="[MASK]",
        **kwargs
    ):
        super().__init__()
        self.token2id = token2id
        self.id2token = {v: k for k, v in self.token2id.items()}
        self.unk_token = unk_token
        self.pad_token = pad_token
        self.cls_token = cls_token
        self.sep_token = sep_token
        self.mask_token = mask_token
        self.unk_id = self.token2id[self.unk_token]
        self.pad_id = self.token2id[self.pad_token]
        self.cls_id = self.token2id[self.cls_token]
        self.sep_id = self.token2id[self.sep_token]
        self.mask_id = self.token2id[self.mask_token]
        self.do_lower_case = do_lower_case

    @classmethod
    def from_file(cls, vocab_file, **kwargs):
        token2id = cls._load_vocab(vocab_file=vocab_file)
        tokenizer = cls(token2id, **kwargs)
        return tokenizer

    @classmethod
    def _load_vocab(cls, vocab_file):
        vocab = {}
        idx = 0
        with open(vocab_file, mode="rt", encoding="utf-8") as fin:
            for line in fin:
                word = line.rstrip("\n")
                vocab[word] = idx
                idx += 1
        return vocab

    def encode(self, text, add_cls=True, add_sep=True, **kwargs):
        tokens, ids, type_ids, attention_mask, offsets = [], [], [], [], []
        if self.do_lower_case:
            text = str(text).lower()
        for idx, char in enumerate(text):
            tokens.append(char)
            ids.append(self.token2id.get(char, self.unk_id))
            type_ids.append(0)
            attention_mask.append(1)
            offsets.append((idx, idx + 1))

        if add_cls:
            tokens.insert(0, self.cls_token)
            ids.insert(0, self.cls_id)
            type_ids.insert(0, 0)
            attention_mask.insert(0, 1)
            offsets.insert(0, (0, 0))

        if add_sep:
            tokens.append(self.sep_token)
            ids.append(self.sep_id)
            type_ids.append(0)
            attention_mask.append(1)
            offsets.append((0, 0))

        encoding = TokenizerEncoding(
            text=text,
            tokens=tokens,
            ids=ids,
            type_ids=type_ids,
            attention_mask=attention_mask,
            offsets=offsets,
        )
        return encoding

    def decode(self, ids, **kwargs):
        tokens = [self.id2token.get(_id, self.unk_token) for _id in ids]
        return tokens

    def token_to_id(self, token, **kwargs):
        if self.do_lower_case:
            token = token.lower()
        return self.token2id.get(token, self.unk_id)

    def id_to_token(self, id, **kwargs):
        return self.id2token.get(id, self.unk_token)



import unittest

from transformers_keras.sequence_classification.dataset import SequenceClassificationDataset


class DatasetTest(unittest.TestCase):
    """Dataset test."""

    def test_sequence_classification_dataset_examples(self):
        print()
        print("====from_jsonl_files")
        d = SequenceClassificationDataset.from_jsonl_files(
            "testdata/sequence_classify.jsonl", vocab_file="testdata/vocab.bert.txt", batch_size=2
        )
        print(next(iter(d)))

        print("====jsonl_to_examples")
        examples = SequenceClassificationDataset.jsonl_to_examples(
            "testdata/sequence_classify.jsonl", vocab_file="testdata/vocab.bert.txt"
        )
        for i in range(2):
            print(examples[i])

        print("====from_examples")
        d = SequenceClassificationDataset.from_examples(examples, batch_size=2)
        print(next(iter(d)))

        print("====examples_to_tfrecord")
        SequenceClassificationDataset.examples_to_tfrecord(
            examples, output_files=["testdata/sequence_classify.tfrecord"]
        )

        print("====from_tfrecord_files")
        d = SequenceClassificationDataset.from_tfrecord_files("testdata/sequence_classify.tfrecord", batch_size=2)
        print(next(iter(d)))


if __name__ == "__main__":
    unittest.main()



import json
import unittest

from transformers_keras.token_classification.dataset import TokenClassificationDataset


class DatasetTest(unittest.TestCase):
    """Dataset tests."""

    def generate_jsonl(self):
        examples = self._read_examples(["testdata/conll.txt"])
        with open("testdata/token_classify.jsonl", mode="wt", encoding="utf-8") as fout:
            for e in examples:
                info = {"features": e.tokens, "labels": e.labels}
                fout.write(json.dumps(info, ensure_ascii=False))
                fout.write("\n")

    def test_dataset(self):
        print("====from_jsonl_files")
        d = TokenClassificationDataset.from_jsonl_files(
            "testdata/token_classify.jsonl",
            vocab_file="testdata/vocab.bert.txt",
            label_vocab_file="testdata/labels.txt",
            batch_size=2,
        )
        print(next(iter(d)))

        print("====jsonl_to_examples")
        examples = TokenClassificationDataset.jsonl_to_examples(
            "testdata/token_classify.jsonl",
            vocab_file="testdata/vocab.bert.txt",
            label_vocab_file="testdata/labels.txt",
        )
        for i in range(2):
            print(examples[i])

        print("====from_examples")
        d = TokenClassificationDataset.from_examples(examples, batch_size=2)
        print(next(iter(d)))

        print("====examples_to_tfrecord")
        TokenClassificationDataset.examples_to_tfrecord(examples, ["testdata/token_classify.tfrecord"])

        print("====from_tfrecord_files")
        d = TokenClassificationDataset.from_tfrecord_files("testdata/token_classify.tfrecord")
        print(next(iter(d)))


if __name__ == "__main__":
    unittest.main()



import tensorflow as tf
from transformers_keras.modeling_bert import BertModel, BertPretrainedModel
from transformers_keras.modeling_utils import choose_activation, shape_list


class BertPredictionHeadTransform(tf.keras.layers.Layer):
    """Transform layer for prediction head."""

    def __init__(self, hidden_size=768, activation="gelu", initializer_range=0.02, epsilon=1e-12, **kwargs):
        super().__init__(**kwargs)
        self.dense = tf.keras.layers.Dense(
            hidden_size,
            kernel_initializer=tf.keras.initializers.TruncatedNormal(stddev=initializer_range),
            name="dense",
        )
        self.activation = choose_activation(activation)
        self.layernorm = tf.keras.layers.LayerNormalization(epsilon=epsilon, name="LayerNorm")

    def call(self, hidden_states, training=None):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.activation(hidden_states)
        hidden_states = self.layernorm(hidden_states)
        return hidden_states


class BertPredictionHead(tf.keras.layers.Layer):
    """Prediction head"""

    def __init__(
        self,
        embedding_table,
        vocab_size=21128,
        hidden_size=768,
        activation="gelu",
        initializer_range=0.02,
        epsilon=1e-12,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.embedding_table = embedding_table
        self.transform = BertPredictionHeadTransform(
            hidden_size=hidden_size,
            activation=activation,
            initializer_range=initializer_range,
            epsilon=epsilon,
            name="transform",
        )
        self.vocab_size = vocab_size
        self.hidden_size = hidden_size

    def build(self, input_shape):
        self.bias = self.add_weight(shape=(self.vocab_size,), initializer="zeros", trainable=True, name="output_bias")
        super().build(input_shape)

    def call(self, hidden_states, training=None):
        hidden_states = self.transform(hidden_states)
        seq_length = shape_list(hidden_states)[1]
        hidden_states = tf.reshape(hidden_states, shape=[-1, self.hidden_size])
        hidden_states = tf.matmul(hidden_states, self.embedding_table, transpose_b=True)
        hidden_states = tf.reshape(hidden_states, shape=[-1, seq_length, self.vocab_size])
        hidden_states = tf.nn.bias_add(value=hidden_states, bias=self.bias)
        return hidden_states


class BertMaskedLanguageModelHead(tf.keras.layers.Layer):
    """Masked language model head for BERT."""

    def __init__(
        self,
        embedding_table,
        vocab_size=21128,
        hidden_size=768,
        activation="gelu",
        initializer_range=0.02,
        epsilon=1e-12,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.prediction = BertPredictionHead(
            embedding_table=embedding_table,
            vocab_size=vocab_size,
            hidden_size=hidden_size,
            activation=activation,
            initializer_range=initializer_range,
            epsilon=epsilon,
            name="predictions",
        )

    def call(self, sequence_output, training=None):
        outputs = self.prediction(sequence_output)
        return outputs


class BertForMaskedLanguageModel(BertPretrainedModel):
    """Bert for masked language model."""

    def __init__(
        self,
        vocab_size=21128,
        max_positions=512,
        hidden_size=768,
        type_vocab_size=2,
        num_layers=6,
        num_attention_heads=8,
        intermediate_size=3072,
        activation="gelu",
        hidden_dropout_rate=0.2,
        attention_dropout_rate=0.1,
        initializer_range=0.02,
        epsilon=1e-12,
        **kwargs
    ):
        input_ids = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="input_ids")
        segment_ids = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="segment_ids")
        attention_mask = tf.keras.layers.Input(shape=(None,), dtype=tf.int32, name="attention_mask")
        bert_model = BertModel(
            vocab_size=vocab_size,
            max_positions=max_positions,
            hidden_size=hidden_size,
            type_vocab_size=type_vocab_size,
            num_layers=num_layers,
            num_attention_heads=num_attention_heads,
            intermediate_size=intermediate_size,
            activation=activation,
            hidden_dropout_rate=hidden_dropout_rate,
            attention_dropout_rate=attention_dropout_rate,
            initializer_range=initializer_range,
            epsilon=epsilon,
            name="bert",
        )
        sequence_output, _, _, _ = bert_model(input_ids, segment_ids, attention_mask)
        embedding_table = bert_model.get_embedding_table()
        mlm = BertMaskedLanguageModelHead(
            embedding_table,
            vocab_size=vocab_size,
            hidden_size=hidden_size,
            activation=activation,
            initializer_range=initializer_range,
            epsilon=epsilon,
            name="cls",
        )
        pred = mlm(sequence_output)
        pred = tf.keras.layers.Lambda(lambda x: x, name="predictions")(pred)
        super().__init__(inputs=[input_ids, segment_ids, attention_mask], outputs=[pred], **kwargs)

        self.epsilon = epsilon

    def train_step(self, data):
        x, y = data
        input_ids, segment_ids, attention_mask = x["input_ids"], x["segment_ids"], x["attention_mask"]
        # masked_ids is golden label, masked_pos is position masking matrix
        masked_pos, y_true = y["masked_pos"], y["masked_ids"]
        with tf.GradientTape() as tape:
            y_pred = self(inputs=[input_ids, segment_ids, attention_mask], training=True)
            loss = self._compute_loss(y_true, y_pred, masked_pos)
        # Compute gradients
        trainable_vars = self.trainable_variables
        # print_variables(trainable_vars)
        gradients = tape.gradient(loss, trainable_vars)
        # Update weights
        self.optimizer.apply_gradients(zip(gradients, trainable_vars))
        # Update metrics (includes the metric that tracks the loss)
        self.compiled_metrics.update_state(y_true, y_pred)
        # Return a dict mapping metric names to current value
        results = {m.name: m.result() for m in self.metrics}
        results.update({"loss": loss})
        return results

    def _compute_loss(self, y_true, y_pred, masked_pos):
        # masked_pos shape: (batch_size, seq_len)
        loss = tf.keras.losses.sparse_categorical_crossentropy(y_true, y_pred, from_logits=True)
        masked_pos = tf.cast(masked_pos, dtype=loss.dtype)
        loss *= masked_pos
        return tf.reduce_sum(loss) / (tf.reduce_sum(masked_pos) + self.epsilon)

    def test_step(self, data):
        x, y = data
        input_ids, segment_ids, attention_mask = x["input_ids"], x["segment_ids"], x["attention_mask"]
        masked_pos, y_true = y["masked_positions"], y["masked_ids"]
        y_pred = self(inputs=[input_ids, segment_ids, attention_mask], training=False)
        loss = self._compute_loss(y_true, y_pred, masked_pos)
        # Update metrics (includes the metric that tracks the loss)
        self.compiled_metrics.update_state(y_true, y_pred)
        # Return a dict mapping metric names to current value
        results = {m.name: m.result() for m in self.metrics}
        results.update({"loss": loss})
        return results



import abc
import json
import logging
import os

import numpy as np
import tensorflow as tf


def zip_weights(model, ckpt, variables_mapping, self_weight_names, **kwargs):
    weights, values = [], []
    used_weights = [w for w in model.trainable_weights if w.name in self_weight_names]
    for w in used_weights:
        var = variables_mapping.get(w.name, None)
        if var is None:
            logging.warning("Model weight: %s not collected in weights mapping.", w.name)
            continue
        v = tf.train.load_variable(ckpt, var)
        if w.name == "bert/nsp/dense/kernel:0":
            v = v.T
        weights.append(w)
        values.append(v)
        if kwargs.get("verbose", True):
            logging.info("Load weight: {:60s} <-- {}".format(w.name, variables_mapping[w.name]))

    mapped_values = zip(weights, values)
    return mapped_values


def parse_pretrained_model_files(pretrained_model_dir):
    config_file, ckpt, vocab = None, None, None
    pretrained_model_dir = os.path.abspath(pretrained_model_dir)
    if not os.path.exists(pretrained_model_dir):
        logging.info("pretrain model dir: {} is not exists.".format(pretrained_model_dir))
        return config_file, ckpt, vocab
    for f in os.listdir(pretrained_model_dir):
        if "config" in str(f) and str(f).endswith(".json"):
            config_file = os.path.join(pretrained_model_dir, f)
        if "vocab" in str(f):
            vocab = os.path.join(pretrained_model_dir, f)
        if "ckpt" in str(f):
            n = ".".join(str(f).split(".")[:-1])
            ckpt = os.path.join(pretrained_model_dir, n)
    return config_file, ckpt, vocab


class AbstractAdapter(abc.ABC):
    """Abstract model weights adapter."""

    def __init__(
        self,
        skip_token_embedding=False,
        skip_position_embedding=False,
        skip_segment_embedding=False,
        skip_embedding_layernorm=False,
        skip_pooler=False,
        skip_mlm=True,
        **kwargs
    ):
        super().__init__()
        self.skip_token_embedding = skip_token_embedding
        self.skip_position_embedding = skip_position_embedding
        self.skip_segment_embedding = skip_segment_embedding
        self.skip_embedding_layernorm = skip_embedding_layernorm
        self.skip_pooler = skip_pooler
        self.skip_mlm = skip_mlm

        logging.info(
            "Adapter skipping config: %s",
            json.dumps(
                {
                    "skip_token_embedding": self.skip_token_embedding,
                    "skip_position_embedding": self.skip_position_embedding,
                    "skip_segment_embedding": self.skip_segment_embedding,
                    "skip_embedding_layernorm": self.skip_embedding_layernorm,
                    "skip_pooler": self.skip_pooler,
                    "skip_mlm": self.skip_mlm,
                }
            ),
        )

    def parse_files(self, model_path, **kwargs):
        config_file, ckpt, vocab = parse_pretrained_model_files(model_path)
        return {
            "config_file": config_file,
            "ckpt": ckpt,
            "vocab_file": vocab,
        }

    @abc.abstractmethod
    def adapte_config(self, config_file, **kwargs):
        raise NotImplementedError()

    @abc.abstractmethod
    def adapte_weights(self, model, ckpt, model_config, use_functional_api=True, **kwargs):
        raise NotImplementedError()

    def _skip_weight(self, mapping, name):
        mapping.pop(name)
        logging.info("Skip load weight: %s", name)

    def _check_weights(self, mapping, model, ckpt, **kwargs):
        self_weights = {w.name: w.numpy() for w in model.trainable_weights}
        for k, v in self_weights.items():
            ckpt_weight = mapping.get(k, None)
            if not ckpt_weight:
                continue
            ckpt_value = tf.train.load_variable(ckpt, ckpt_weight)
            if ckpt_value is None:
                logging.warning("ckpt value is None of key: %s", ckpt_weight)
            assert np.allclose(v, ckpt_value)
        logging.info("All weights value are checked.")



import unittest

from transformers_keras.lm.mlm_dataset import MaskedLanguageModelDataset


class MaskedLanguageModelDatasetTest(unittest.TestCase):
    """Test dataset for masked language model"""

    def test_mlm_dataset(self):
        print("\n=====jsonl_to_examples")
        examples = MaskedLanguageModelDataset.jsonl_to_examples(
            input_files=["testdata/mlm.jsonl"],
            vocab_file="testdata/vocab.bert.txt",
        )
        for idx, e in enumerate(examples):
            print("{} -> {}".format(idx, e))

        print("=====from_examples")
        d = MaskedLanguageModelDataset.from_examples(examples, batch_size=2)
        print(next(iter(d)))

        print("=====examples_to_tfrecord")
        MaskedLanguageModelDataset.examples_to_tfrecord(examples, output_files=["testdata/mlm.tfrecord"])
        print("=====from_tfrecord_files")
        d = MaskedLanguageModelDataset.from_tfrecord_files(input_files=["testdata/mlm.tfrecord"], batch_size=2)
        print(next(iter(d)))


if __name__ == "__main__":
    unittest.main()



"""Opinion Term Extraction and Classification"""

from transformers_keras.question_answering.models import AlbertForQuestionAnsweringX, BertForQuestionAnsweringX


class BertForOpinionTermExtractionAndClassification(BertForQuestionAnsweringX):
    """Use QA model to extract opinion term and classify sentiment polarity."""

    pass


class AlbertForOpinionTermExtractionAndClassification(AlbertForQuestionAnsweringX):
    """Use QA model to extract opinion term and classify sentiment polarity."""

    pass



import json
import logging

import tensorflow as tf

from .abstract_adapter import AbstractAdapter, zip_weights


class AlbertAdapter(AbstractAdapter):
    def __init__(self, skip_embedding_mapping_in=False, **kwargs):
        super().__init__(**kwargs)
        self.skip_embedding_mapping_in = skip_embedding_mapping_in

    def adapte_config(self, config_file, **kwargs):
        with open(config_file, mode="rt", encoding="utf8") as fin:
            config = json.load(fin)

        model_config = {
            "vocab_size": config["vocab_size"],
            "max_positions": config["max_position_embeddings"],
            "embedding_size": config["embedding_size"],
            "type_vocab_size": config["type_vocab_size"],
            "num_layers": config["num_hidden_layers"],
            "num_groups": config["num_hidden_groups"],
            "num_layers_each_group": config["inner_group_num"],
            "hidden_size": config["hidden_size"],
            "num_attention_heads": config["num_attention_heads"],
            "intermediate_size": config["intermediate_size"],
            "activation": config["hidden_act"],
            "hidden_dropout_rate": config["hidden_dropout_prob"],
            "attention_dropout_rate": config["attention_probs_dropout_prob"],
            "initializer_range": config["initializer_range"],
        }
        return model_config

    def adapte_weights(self, model, ckpt, model_config, use_functional_api=True, **kwargs):
        ckpt_weight_names = [x[0] for x in tf.train.list_variables(ckpt)]
        self_weight_names = set([x.name for x in model.trainable_weights])
        weight_mapping = {}
        albert_weight_mapping = self._adapte_albert_weights(
            model, ckpt, model_config, use_functional_api=use_functional_api, **kwargs
        )
        weight_mapping.update(albert_weight_mapping)

        mlm_weight_mapping = self._adapte_mlm_weights(model, ckpt, use_functional_api=use_functional_api, **kwargs)
        weight_mapping.update(mlm_weight_mapping)

        sop_weight_mapping = self._adapte_sop_weights(model, ckpt, use_functional_api=use_functional_api, **kwargs)
        weight_mapping.update(sop_weight_mapping)

        if kwargs.get("check_weights", False):
            self._check_weights(weight_mapping, model, ckpt)

    def _adapte_mlm_weights(self, model, ckpt, use_functional_api=True, **kwargs):
        with_mlm = kwargs.get("with_mlm", False)
        if not with_mlm:
            logging.info("Skipping to adapte weights for MLM due to option `with_mlm` set to `False`")
            return {}
        self_mlm_prefix = "cls/predictions" if use_functional_api else model.name + "/cls/predictions"
        ckpt_mlm_prefix = kwargs.get("ckpt_mlm_prefix", "cls/predictions")
        logging.info("Adapting MLM weights, using model weight prefix: %s", self_mlm_prefix)
        logging.info("Adapting MLM weights, using  ckpt weight prefix: %s", ckpt_mlm_prefix)
        ckpt_weight_names = [x for (x, _) in tf.train.list_variables(ckpt) if str(x).startswith(ckpt_mlm_prefix)]
        self_weight_names = set([x.name for x in model.trainable_weights if str(x.name).startswith(self_mlm_prefix)])
        mapping = {}
        for w in ckpt_weight_names:
            if str(w).startswith(ckpt_mlm_prefix):
                mw = self_mlm_prefix + w[len(ckpt_mlm_prefix) :] + ":0"
                if mw not in self_weight_names:
                    logging.warning("weight: %s not in model weights", mw)
                    continue
                mapping[mw] = w

        # zip weight names and values
        zipped_weights = zip_weights(model, ckpt, mapping, self_weight_names, **kwargs)
        # set values to weights
        tf.keras.backend.batch_set_value(zipped_weights)

        return mapping

    def _adapte_sop_weights(self, model, ckpt, use_functional_api=True, **kwargs):
        with_sop = kwargs.get("with_sop", False)
        if not with_sop:
            logging.info("Skipping to adapte weights for SOP due to option `with_sop` set to `False`")
            return {}
        self_sop_prefix = "cls/seq_relationship" if use_functional_api else model.name + "/cls/seq_relationship"
        ckpt_sop_prefix = kwargs.get("ckpt_sop_prefix", "cls/seq_relationship")
        logging.info("Adapting SOP weights, using model weight prefix: %s", self_sop_prefix)
        logging.info("Adapting SOP weights, using  ckpt weight prefix: %s", ckpt_sop_prefix)
        ckpt_weight_names = [x for (x, _) in tf.train.list_variables(ckpt) if str(x).startswith(ckpt_sop_prefix)]
        self_weight_names = set([x.name for x in model.trainable_weights if str(x.name).startswith(self_sop_prefix)])
        mapping = {}
        for w in ckpt_weight_names:
            # TODO skip unused variables
            if str(w).startswith(ckpt_sop_prefix):
                mw = self_sop_prefix + w.lstrip(ckpt_sop_prefix) + ":0"
                if mw not in self_weight_names:
                    logging.warning("weight: %s not in model weights", mw)
                    continue
                mapping[mw] = w
        # zip weight names and values
        zipped_weights = zip_weights(model, ckpt, mapping, self_weight_names, **kwargs)
        # set values to weights
        tf.keras.backend.batch_set_value(zipped_weights)

        return mapping

    def _adapte_albert_weights(self, model, ckpt, model_config, use_functional_api=True, **kwargs):
        mapping = {}
        self_albert_prefix = (
            model.albert_model.name if use_functional_api else model.name + "/" + model.albert_model.name
        )
        ckpt_albert_prefix = kwargs.get("ckpt_albert_prefix", "bert")
        logging.info("Adapting albert weights, using model weight prefix: %s", self_albert_prefix)
        logging.info("Adapting albert weights, using  ckpt weight prefix: %s", ckpt_albert_prefix)
        self_weight_names = set([x.name for x in model.trainable_weights if str(x.name).startswith(self_albert_prefix)])
        ckpt_weight_names = [x for (x, _) in tf.train.list_variables(ckpt) if str(x).startswith(ckpt_albert_prefix)]

        embedding_mapping = self._adapte_embedding_weights(
            model, ckpt, self_weight_names, ckpt_weight_names, use_functional_api=use_functional_api, **kwargs
        )
        mapping.update(embedding_mapping)

        encoder_mapping = self._adapte_encoder_weights(
            model,
            ckpt,
            self_weight_names,
            ckpt_weight_names,
            num_groups=model_config["num_groups"],
            num_layers_each_group=model_config["num_layers_each_group"],
            **kwargs
        )
        mapping.update(encoder_mapping)

        # skip weights
        self_albert_prefix = (
            model.albert_model.name if use_functional_api else model.name + "/" + model.albert_model.name
        )
        self._skip_weights(mapping, self_albert_prefix)

        # zip weight names and values
        zipped_weights = zip_weights(model, ckpt, mapping, self_weight_names, **kwargs)
        # set values to weights
        tf.keras.backend.batch_set_value(zipped_weights)

        # check weights
        if kwargs.get("check_weights", False):
            self._check_weights(mapping, model, ckpt, **kwargs)

        return mapping

    def _adapte_embedding_weights(
        self, model, ckpt, self_weight_names, ckpt_weight_names, use_functional_api=True, **kwargs
    ):
        model_prefix = model.albert_model.name if use_functional_api else model.name + "/" + model.albert_model.name
        ckpt_prefix = kwargs.get("ckpt_albert_prefix", "bert")
        mapping = {}
        for w in ckpt_weight_names:
            # mapping embedding weights
            if any(x in w for x in ["embeddings", "pooler"]):
                mw = model_prefix + w.lstrip(ckpt_prefix) + ":0"
                if mw not in self_weight_names:
                    logging.warning("weight: %s not in model weights", mw)
                    continue
                mapping[mw] = w
            # mapping embedding mapin weights
            if "embedding_hidden_mapping_in/kernel" in w:
                mw = model_prefix + "/encoder/embedding_mapping/kernel:0"
                if mw not in self_weight_names:
                    logging.warning("weight: %s not in model weights", mw)
                    continue
                mapping[mw] = w
            if "embedding_hidden_mapping_in/bias" in w:
                mw = model_prefix + "/encoder/embedding_mapping/bias:0"
                if mw not in self_weight_names:
                    logging.warning("weight: %s not in model weights", mw)
                    continue
                mapping[mw] = w
        return mapping

    def _adapte_encoder_weights(
        self,
        model,
        ckpt,
        self_weight_names,
        ckpt_weight_names,
        num_groups,
        num_layers_each_group,
        use_functional_api=True,
        **kwargs
    ):
        model_prefix = model.albert_model.name if use_functional_api else model.name + "/" + model.albert_model.name
        ckpt_prefix = kwargs.get("ckpt_albert_prefix", "bert")
        mapping = {}
        for group in range(num_groups):
            for layer in range(num_layers_each_group):
                k_prefix = "{}/encoder/group_{}/layer_{}/".format(model_prefix, group, layer)
                v_prefix = "{}/encoder/transformer/group_{}/inner_group_{}/".format(ckpt_prefix, group, layer)
                # attention
                for n in ["query", "key", "value"]:
                    for x in ["kernel", "bias"]:
                        k = k_prefix + "attention/{}/{}:0".format(n, x)
                        v = v_prefix + "attention_1/self/{}/{}".format(n, x)
                        if k not in self_weight_names:
                            logging.warning("weight: %s not in model weights", k)
                            continue
                        mapping[k] = v

                # attention dense
                for n in ["kernel", "bias"]:
                    k = k_prefix + "attention/dense/{}:0".format(n)
                    v = v_prefix + "attention_1/output/dense/{}".format(n)
                    if k not in self_weight_names:
                        logging.warning("weight: %s not in model weights", k)
                        continue
                    mapping[k] = v

                for n in ["gamma", "beta"]:
                    # attention layer norm
                    k = k_prefix + "attention/layer_norm/{}:0".format(n)
                    v = v_prefix + "LayerNorm/{}".format(n)
                    if k not in self_weight_names:
                        logging.warning("weight: %s not in model weights", k)
                        continue
                    mapping[k] = v
                    # albert encoder layer norm
                    k = k_prefix + "layer_norm/{}:0".format(n)
                    v = v_prefix + "LayerNorm_1/{}".format(n)
                    if k not in self_weight_names:
                        logging.warning("weight: %s not in model weights", k)
                        continue
                    mapping[k] = v

                for n in ["kernel", "bias"]:
                    # intermediate
                    k = k_prefix + "ffn/{}:0".format(n)
                    v = v_prefix + "ffn_1/intermediate/dense/{}".format(n)
                    if k not in self_weight_names:
                        logging.warning("weight: %s not in model weights", k)
                        continue
                    mapping[k] = v
                    # dense
                    k = k_prefix + "ffn_output/{}:0".format(n)
                    v = v_prefix + "ffn_1/intermediate/output/dense/{}".format(n)
                    if k not in self_weight_names:
                        logging.warning("weight: %s not in model weights", k)
                        continue
                    mapping[k] = v
        return mapping

    def _skip_weights(self, mapping, model_prefix):
        if self.skip_token_embedding:
            self._skip_weight(mapping, model_prefix + "/embeddings/word_embeddings:0")
        if self.skip_position_embedding:
            self._skip_weight(mapping, model_prefix + "/embeddings/position_embeddings:0")
        if self.skip_segment_embedding:
            self._skip_weight(mapping, model_prefix + "/embeddings/token_type_embeddings:0")
        if self.skip_embedding_layernorm:
            self._skip_weight(mapping, model_prefix + "/embeddings/LayerNorm/gamma:0")
            self._skip_weight(mapping, model_prefix + "/embeddings/LayerNorm/beta:0")
        if self.skip_pooler:
            self._skip_weight(mapping, model_prefix + "/pooler/dense/kernel:0")
            self._skip_weight(mapping, model_prefix + "/pooler/dense/bias:0")



import tensorflow as tf


class Distiller(tf.keras.Model):
    """Model distiller."""

    def __init__(self, teacher: tf.keras.Model, student: tf.keras.Model, **kwargs):
        super().__init__(inputs=student.inputs, outputs=student.outputs, **kwargs)
        self.teacher = teacher
        self.teacher.trainable = False
        self.student = student

    def compile(self, student_loss, distill_loss, optimizer, metrics=None, alpha=0.1, temperature=3, **kwargs):
        super().compile(optimizer=optimizer, metrics=metrics, **kwargs)
        self.student_loss = student_loss
        self.distill_loss = distill_loss
        self.alpha = alpha
        self.temperature = temperature

    def train_step(self, data):
        # Unpack data
        x, y = data
        teacher_predictions = self.teacher(x, training=False)
        with tf.GradientTape() as tape:
            # Forward pass of student
            student_predictions = self.student(x, training=True)
            # Compute losses
            _student_loss = self.student_loss(y, student_predictions)
            _distill_loss = self.distill_loss(
                tf.nn.softmax(teacher_predictions / self.temperature, axis=1),
                tf.nn.softmax(student_predictions / self.temperature, axis=1),
            )
            loss = self.alpha * _student_loss + (1 - self.alpha) * _distill_loss

        # Compute gradients
        trainable_vars = self.student.trainable_variables
        gradients = tape.gradient(loss, trainable_vars)
        # Update weights
        self.optimizer.apply_gradients(zip(gradients, trainable_vars))
        # Update the metrics configured in `compile()`.
        self.compiled_metrics.update_state(y, student_predictions)
        # Return a dict of performance
        results = {m.name: m.result() for m in self.metrics}
        results.update({"student_loss": _student_loss, "distill_loss": _distill_loss})
        return results

    def test_step(self, data):
        # Unpack the data
        x, y = data
        # Compute predictions
        y_prediction = self.student(x, training=False)
        # Calculate the loss
        _student_loss = self.student_loss(y, y_prediction)
        # Update the metrics.
        self.compiled_metrics.update_state(y, y_prediction)
        # Return a dict of performance
        results = {m.name: m.result() for m in self.metrics}
        results.update({"student_loss": _student_loss})
        return results



import abc
import logging
import random
from collections import namedtuple

import tensorflow as tf
from tokenizers import BertWordPieceTokenizer
from transformers_keras.common.abc_dataset import AbstractDataset

MaskedLangueModelExample = namedtuple(
    "MaskedLanguageModelExample",
    ["tokens", "input_ids", "segment_ids", "attention_mask", "masked_ids", "masked_pos"],
)


class WholeWordMask(abc.ABC):
    """Default masking strategy from BERT."""

    def __init__(self, change_prob=0.15, mask_prob=0.8, rand_prob=0.1, keep_prob=0.1, max_predictions=20, **kwargs):
        self.change_prob = change_prob
        self.mask_prob = mask_prob / (mask_prob + rand_prob + keep_prob)
        self.rand_prob = rand_prob / (mask_prob + rand_prob + keep_prob)
        self.keep_prob = keep_prob / (mask_prob + rand_prob + keep_prob)
        self.max_predictions = max_predictions

    def __call__(self, sequence, tokenizer: BertWordPieceTokenizer, vocabs, max_sequence_length=512, **kwargs):
        encoding = tokenizer.encode(sequence, add_special_tokens=False)
        tokens = encoding.tokens
        tokens = self._truncate_sequence(tokens, max_sequence_length - 2)
        if not tokens:
            return None
        num_to_predict = min(self.max_predictions, max(1, round(self.change_prob * len(tokens))))
        cand_indexes = self._collect_candidates(tokens)
        # copy original tokens
        masked_tokens = [x for x in tokens]
        masked_indexes = [0] * len(tokens)
        for piece_indexes in cand_indexes:
            if sum(masked_indexes) >= num_to_predict:
                break
            if sum(masked_indexes) + len(piece_indexes) > num_to_predict:
                continue
            if any(masked_indexes[idx] == 1 for idx in piece_indexes):
                continue
            for index in piece_indexes:
                masked_indexes[index] = 1
                masked_tokens[index] = self._masking_tokens(index, tokens, vocabs)

        # add special tokens
        tokens = ["[CLS]"] + tokens + ["[SEP]"]
        masked_tokens = ["[CLS]"] + masked_tokens + ["[SEP]"]
        masked_indexes = [0] + masked_indexes + [0]
        assert len(tokens) == len(masked_tokens) == len(masked_indexes)

        example = MaskedLangueModelExample(
            tokens=masked_tokens,
            input_ids=[tokenizer.token_to_id(x) for x in masked_tokens],
            segment_ids=[0] * len(masked_tokens),
            attention_mask=[1] * len(masked_tokens),
            masked_ids=[tokenizer.token_to_id(x) for x in tokens],
            masked_pos=masked_indexes,
        )
        return example

    def _masking_tokens(self, index, tokens, vocabs, **kwargs):
        # 80% of the time, replace with [MASK]
        if random.random() < self.mask_prob:
            return "[MASK]"
        # 10% of the time, keep original
        p = self.rand_prob / (self.rand_prob + self.keep_prob)
        if random.random() < p:
            return tokens[index]
        # 10% of the time, replace with random word
        masked_token = vocabs[random.randint(0, len(vocabs) - 1)]
        return masked_token

    def _collect_candidates(self, tokens):
        cand_indexes = [[]]
        for idx, token in enumerate(tokens):
            if cand_indexes and token.startswith("##"):
                cand_indexes[-1].append(idx)
                continue
            cand_indexes.append([idx])
        random.shuffle(cand_indexes)
        return cand_indexes

    def _truncate_sequence(self, tokens, max_tokens=512, **kwargs):
        while len(tokens) > max_tokens:
            if len(tokens) > max_tokens:
                tokens.pop(0)
                # truncate whole world
                while tokens and tokens[0].startswith("##"):
                    tokens.pop(0)
            if len(tokens) > max_tokens:
                while tokens and tokens[-1].startswith("##"):
                    tokens.pop()
                if tokens:
                    tokens.pop()
        return tokens


class MaskedLanguageModelDataset(AbstractDataset):
    """Dataset builder for masked language model."""

    @classmethod
    def _filter(cls, dataset, max_sequence_length=512, **kwargs):
        dataset = dataset.filter(
            lambda a, b, c, x, y: tf.size(a) <= max_sequence_length,
        )
        return dataset

    @classmethod
    def _bucket_padding(
        cls,
        dataset,
        batch_size=32,
        pad_id=0,
        bucket_boundaries=[50, 100, 150, 200, 250, 200, 250, 400, 450],
        bucket_batch_sizes=None,
        drop_remainder=False,
        **kwargs
    ):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = cls._bucketing(
            dataset,
            element_length_func=lambda a, b, c, x, y: tf.size(a),
            padded_shapes=([None,], [None,], [None,], [None,], [None,]),
            padding_values=(pad_id, pad_id, pad_id, pad_id, pad_id),
            batch_size=batch_size,
            bucket_boundaries=bucket_boundaries,
            bucket_batch_sizes=bucket_batch_sizes,
            drop_remainder=drop_remainder,
            **kwargs,
        )
        # fmt: on
        return dataset

    @classmethod
    def _batch_padding(cls, dataset, batch_size=32, pad_id=0, drop_remainder=False, **kwargs):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([None,], [None,], [None,], [None,], [None,]),
            padding_values=(pad_id, pad_id, pad_id, pad_id, pad_id),
            drop_remainder=drop_remainder,
        )
        # fmt: on
        return dataset

    @classmethod
    def _fixed_padding(cls, dataset, batch_size=32, pad_id=0, max_sequence_length=512, drop_remainder=False, **kwargs):
        pad_id = tf.constant(pad_id, dtype=tf.int32)
        maxlen = tf.constant(max_sequence_length, dtype=tf.int32)
        # fmt: off
        dataset = dataset.padded_batch(
            batch_size,
            padded_shapes=([maxlen,], [maxlen,], [maxlen,], [maxlen,], [maxlen,]),
            padding_values=(pad_id, pad_id, pad_id, pad_id, pad_id),
            drop_remainder=drop_remainder,
        )
        # fmt: on
        return dataset

    @classmethod
    def _zip_dataset(cls, examples, **kwargs):
        def _to_dataset(x, dtype=tf.int32):
            x = tf.ragged.constant(x, dtype=dtype)
            d = tf.data.Dataset.from_tensor_slices(x)
            d = d.map(lambda x: x)
            return d

        dataset = tf.data.Dataset.zip(
            (
                _to_dataset([e.input_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.segment_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.attention_mask for e in examples], dtype=tf.int32),
                _to_dataset([e.masked_ids for e in examples], dtype=tf.int32),
                _to_dataset([e.masked_pos for e in examples], dtype=tf.int32),
            )
        )
        return dataset

    @classmethod
    def _to_dict(cls, dataset, **kwargs):
        dataset = dataset.map(
            lambda a, b, c, x, y: (
                {"input_ids": a, "segment_ids": b, "attention_mask": c},
                {"masked_ids": x, "masked_pos": y},
            ),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _example_to_tfrecord(cls, example, **kwargs):
        feature = {
            "input_ids": cls._int64_feature([int(x) for x in example.input_ids]),
            "segment_ids": cls._int64_feature([int(x) for x in example.segment_ids]),
            "attention_mask": cls._int64_feature([int(x) for x in example.attention_mask]),
            "masked_ids": cls._int64_feature([int(x) for x in example.masked_ids]),
            "masked_pos": cls._int64_feature([int(x) for x in example.masked_pos]),
        }
        return tf.train.Example(features=tf.train.Features(feature=feature))

    @classmethod
    def _parse_tfrecord(cls, dataset, **kwargs):
        features = {
            "input_ids": tf.io.VarLenFeature(tf.int64),
            "segment_ids": tf.io.VarLenFeature(tf.int64),
            "attention_mask": tf.io.VarLenFeature(tf.int64),
            "masked_ids": tf.io.VarLenFeature(tf.int64),
            "masked_pos": tf.io.VarLenFeature(tf.int64),
        }
        dataset = dataset.map(
            lambda x: tf.io.parse_example(x, features),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        dataset = dataset.map(
            lambda x: (
                tf.cast(tf.sparse.to_dense(x["input_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["segment_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["attention_mask"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["masked_ids"]), tf.int32),
                tf.cast(tf.sparse.to_dense(x["masked_pos"]), tf.int32),
            ),
            num_parallel_calls=cls.AUTOTUNE,
        ).prefetch(cls.AUTOTUNE)
        return dataset

    @classmethod
    def _parse_jsonl(cls, instances, tokenizer=None, vocab_file=None, mask_strategy=None, **kwargs):
        assert tokenizer or vocab_file, "`tokenizer` or `vocab_file` must be provided."
        if tokenizer is None:
            tokenizer = BertWordPieceTokenizer.from_file(
                vocab_file,
                lowercase=kwargs.get("do_lower_case", True),
            )
        if mask_strategy is None:
            logging.info("Using default whole word mask strategy...")
            mask_strategy = WholeWordMask(
                change_prob=kwargs.get("change_prob", 0.15),
                mask_prob=kwargs.get("mask_prob", 0.8),
                rand_prob=kwargs.get("rand_prob", 0.1),
                keep_prob=kwargs.get("keep_prob", 0.1),
                max_predictions=kwargs.get("max_predictions", 20),
            )

        examples = []
        vocabs = list(tokenizer.get_vocab().keys())
        for instance in instances:
            sequence = instance.get(kwargs.get("sequence_key", "sequence"), "")
            if not sequence:
                continue
            e = mask_strategy(sequence, tokenizer, vocabs, **kwargs)
            if not e:
                continue
            examples.append(e)
        logging.info("Collected %d examples in total.", len(examples))
        return examples



import unittest

from transformers_keras.sentence_embedding.simcse_dataset import (
    HardNegativeSimCSEDataset,
    SupervisedSimCSEDataset,
    UnsupervisedSimCSEDataset,
)


class SimCSEDatasetTest(unittest.TestCase):
    """SimCSE dataset tests."""

    def test_unsup_dataset(self):
        print("====from_jsonl_files\n")
        d = UnsupervisedSimCSEDataset.from_jsonl_files(
            "testdata/simcse.jsonl", vocab_file="testdata/vocab.bert.txt", batch_size=2
        )
        print(next(iter(d)))

        print("====jsonl_to_examples\n")
        examples = UnsupervisedSimCSEDataset.jsonl_to_examples(
            "testdata/simcse.jsonl", vocab_file="testdata/vocab.bert.txt"
        )
        for i in range(2):
            print(examples[i])

        print("====from_examples\n")
        d = UnsupervisedSimCSEDataset.from_examples(examples, batch_size=2)
        print(next(iter(d)))

        print("====examples_to_tfrecord\n")
        UnsupervisedSimCSEDataset.examples_to_tfrecord(examples, "testdata/simcse.unsup.tfrecord")

        print("====from_tfrecord_files\n")
        d = UnsupervisedSimCSEDataset.from_tfrecord_files("testdata/simcse.unsup.tfrecord", batch_size=2)
        print(next(iter(d)))

    def test_supervised_dataset(self):
        print("====from_jsonl_files\n")
        d = SupervisedSimCSEDataset.from_jsonl_files(
            "testdata/simcse.jsonl", vocab_file="testdata/vocab.bert.txt", batch_size=2
        )
        print(next(iter(d)))

        print("====jsonl_to_examples\n")
        examples = SupervisedSimCSEDataset.jsonl_to_examples(
            "testdata/simcse.jsonl", vocab_file="testdata/vocab.bert.txt"
        )
        for i in range(2):
            print(examples[i])

        print("====from_examples\n")
        d = SupervisedSimCSEDataset.from_examples(examples, batch_size=2)
        print(next(iter(d)))

        print("====examples_to_tfrecord\n")
        SupervisedSimCSEDataset.examples_to_tfrecord(examples, "testdata/simcse.sup.tfrecord")

        print("====from_tfrecord_files\n")
        d = SupervisedSimCSEDataset.from_tfrecord_files("testdata/simcse.sup.tfrecord", batch_size=2)
        print(next(iter(d)))

    def test_hardneg_dataset(self):
        print("====from_jsonl_files\n")
        d = HardNegativeSimCSEDataset.from_jsonl_files(
            "testdata/simcse.jsonl", vocab_file="testdata/vocab.bert.txt", batch_size=2
        )
        print(next(iter(d)))

        print("====jsonl_to_examples\n")
        examples = HardNegativeSimCSEDataset.jsonl_to_examples(
            "testdata/simcse.jsonl", vocab_file="testdata/vocab.bert.txt"
        )
        for i in range(2):
            print(examples[i])

        print("====from_examples\n")
        d = HardNegativeSimCSEDataset.from_examples(examples, batch_size=2)
        print(next(iter(d)))

        print("====examples_to_tfrecord\n")
        HardNegativeSimCSEDataset.examples_to_tfrecord(examples, "testdata/simcse.hardneg.tfrecord")

        print("====from_tfrecord_files\n")
        d = HardNegativeSimCSEDataset.from_tfrecord_files("testdata/simcse.hardneg.tfrecord", batch_size=2)
        print(next(iter(d)))


if __name__ == "__main__":
    unittest.main()



import logging
import os
import unittest

import tensorflow as tf
from tokenizers import BertWordPieceTokenizer
from transformers_keras.lm.mlm import BertForMaskedLanguageModel
from transformers_keras.lm.mlm_dataset import MaskedLanguageModelDataset

BERT_PATH = os.path.join(os.environ["CHINESE_BERT_PATH"], "chinese_roberta_wwm_ext_L-3_H-768_A-12")


class MaskedLanguageModelTest(tf.test.TestCase):
    """Tests for masked language model."""

    def test_bert_for_mlm(self):
        m = BertForMaskedLanguageModel.from_pretrained(BERT_PATH, with_mlm=True)
        m.summary()

    def test_bert_for_mlm_train(self):
        d = MaskedLanguageModelDataset.from_tfrecord_files(
            input_files=["testdata/mlm.tfrecord"], batch_size=2, repeat=100
        )
        m = BertForMaskedLanguageModel.from_pretrained(BERT_PATH, with_mlm=True)
        m.summary()
        m.compile(optimizer="adam")
        m.fit(d)

        m.save("models/bert-for-mlm/1")

    def test_bert_for_mlm_load(self):
        m = tf.saved_model.load("models/bert-for-mlm/1")
        m = m.signatures["serving_default"]
        print(m.structured_input_signature)
        print(m.structured_outputs)

    def _build_model_inputs(self):
        vocab_path = os.path.join(os.environ["GOOGLE_BERT_PATH"], "uncased_L-6_H-768_A-12", "vocab.txt")
        tokenizer = BertWordPieceTokenizer(vocab_path)
        encoding = tokenizer.encode("I love NLP, Neural [MASK] Processing is amazing!")
        logging.info("   ids: %s", encoding.ids)
        logging.info("tokens: %s", encoding.tokens)
        input_ids = tf.constant([encoding.ids], dtype=tf.int32, shape=(1, len(encoding.ids)))
        segment_ids = tf.constant([encoding.type_ids], dtype=tf.int32, shape=(1, len(encoding.type_ids)))
        attention_mask = tf.constant([encoding.attention_mask], dtype=tf.int32, shape=(1, len(encoding.attention_mask)))
        return input_ids, segment_ids, attention_mask

    def _build_masked_lm_from_huggingface(self):
        from transformers import TFBertForMaskedLM

        model = TFBertForMaskedLM.from_pretrained(
            os.path.join(os.environ["PYTORCH_MODEL_PATH"], "bert_uncased_L-6_H-768_A-12-pytorch"),
            from_pt=True,
        )
        return model

    def _build_masked_lm(self):
        model = BertForMaskedLanguageModel.from_pretrained(
            os.path.join(os.environ["GOOGLE_BERT_PATH"], "uncased_L-6_H-768_A-12"),
            with_mlm=True,
            override_params={"epsilon": 1e-12},
        )
        return model

    def test_compare_with_huggingface(self):
        # TODO
        my_model = self._build_masked_lm()
        hg_model = self._build_masked_lm_from_huggingface()

        input_ids, segment_ids, attention_mask = self._build_model_inputs()

        def _compare_embedding_outputs():
            a = my_model.bert_model.bert_embedding(input_ids, segment_ids)
            embedding = hg_model.get_layer("bert").embeddings
            b = embedding(input_ids=input_ids, token_type_ids=segment_ids, position_ids=None, training=False)
            self.assertAllClose(a, b)

        def _compare_hidden_states():
            a_sequence, a_pooled, a_hidden, a_attn = my_model.bert_model(
                input_ids, segment_ids, attention_mask, training=False
            )
            b_sequence, b_pooled, b_hidden, b_attn = hg_model.get_layer("bert")(
                input_ids=input_ids,
                token_type_ids=segment_ids,
                attention_mask=attention_mask,
                return_dict=False,
                output_attentions=True,
                output_hidden_states=True,
            )
            b_attn = tf.transpose(b_attn, [1, 0, 2, 3, 4])
            self.assertAllClose(a_attn, b_attn, rtol=7e-4, atol=9e-5)
            self.assertAllClose(a_sequence, b_sequence, rtol=3e-1, atol=7e-5)

        def _compare_final_outputs():
            a_pred = my_model(inputs=[input_ids, segment_ids, attention_mask])
            outputs = hg_model(
                input_ids=input_ids, token_type_ids=segment_ids, attention_mask=attention_mask, return_dict=True
            )
            b_pred = outputs["logits"]
            self.assertAllClose(a_pred, b_pred, rtol=7e-1, atol=4e-4)

        _compare_embedding_outputs()
        logging.info("Passed comparing embedding outputs...")
        _compare_hidden_states()
        logging.info("Passed comparing hidden states...")
        _compare_final_outputs()
        logging.info("Passed comparing final outputs...")


if __name__ == "__main__":
    unittest.main()



import logging
from typing import List

import numpy as np
import tensorflow as tf
from transformers_keras.common.metrics import ExactMatch, F1ForSequence
from transformers_keras.sentiment_analysis.dataset import AspectTermExtractionDataset, AspectTermExtractionExample


class BaseMetricForAspectTermExtraction(tf.keras.callbacks.Callback):
    """Base metric for ATE"""

    @classmethod
    def from_jsonl_files(cls, input_files, limit=None, **kwargs):
        examples = AspectTermExtractionDataset.jsonl_to_examples(input_files, **kwargs)
        if limit is not None and limit > 0:
            examples = examples[:limit]
        return cls(examples, **kwargs)

    def __init__(self, examples: List[AspectTermExtractionExample], **kwargs):
        super().__init__()
        self.examples = examples
        self.dataset = AspectTermExtractionDataset.from_examples(self.examples, **kwargs)

    def on_epoch_end(self, epoch, logs=None):
        outputs = self.model.predict(self.dataset)
        all_start_ids, all_end_ids = outputs[0], outputs[1]
        pred_spans_list, gold_spans_list = [], []
        for start_ids, end_ids, example in zip(all_start_ids, all_end_ids, self.examples):
            start_ids, end_ids = np.argmax(start_ids, axis=-1), np.argmax(end_ids, axis=-1)
            pred_spans_list.append(self._decode_spans(start_ids.tolist(), end_ids.tolist(), example))
            gold_spans_list.append(self._decode_spans(example.start_ids, example.end_ids, example))
        self._compute_metric(gold_spans_list, pred_spans_list, epoch=epoch)

    def _compute_metric(self, gold_spans_list, pred_spans_list, epoch=0):
        pass

    def _decode_spans(self, start_ids, end_ids, example):
        spans = []
        span = []
        for idx, (start, end) in enumerate(zip(start_ids, end_ids)):
            if start == 1:
                span = [idx]
            if span and end == 1:
                span.append(idx)
                spans.append("".join([str(x).lstrip("##") for x in example.tokens[span[0] : span[1] + 1]]))
                span = []
        return spans


class ExactMatchForAspectTermExtraction(BaseMetricForAspectTermExtraction):
    """EM for ATE"""

    def __init__(self, examples: List[AspectTermExtractionExample], **kwargs) -> None:
        super().__init__(examples, **kwargs)
        self.em = ExactMatch()

    def _compute_metric(self, gold_spans_list, pred_spans_list, epoch=0):
        score = self.em(gold_spans_list, pred_spans_list, dim=2)
        logging.info("Epoch: %d, EM: %.6f", epoch, score)
        tf.summary.scalar("EM", score, description="EM score")


class F1ForAspectTermExtraction(BaseMetricForAspectTermExtraction):
    """F1 for ATE"""

    def __init__(self, examples: List[AspectTermExtractionExample], split_whitespace=False, **kwargs):
        super().__init__(examples, **kwargs)
        self.split_whitespace = split_whitespace
        self.f1 = F1ForSequence()

    def _compute_metric(self, gold_spans_list, pred_spans_list, epoch=0):
        score = self.f1(gold_spans_list, pred_spans_list, dim=2, split_whitespace=self.split_whitespace)
        logging.info("Epoch: %d, F1: %.6f", epoch, score)
        tf.summary.scalar("F1", score, description="F1 score")
